try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from matplotlib import pyplot
import numpy as np
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_base import OMFITtree

__all__ = [
    'OMFITeliteGamma',
    'OMFITeliteEigenfunction',
    'OMFITelite2DEigenfunction',
    'OMFITelitefun2d',
    'OMFITelitextopsi',
    'OMFITeliteAggregated',
    'OMFITeqdat',
    'OMFITbalstab',
    'OMFIToutbal',
]


class OMFITeliteGamma(SortedDict, OMFITascii):
    """ELITE growth rates data file (Gamma file)"""

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        self.clear()
        with open(self.filename, 'r') as f:
            lines = f.readlines()
        for line in lines[1:]:
            tmp = line.split()
            nn = int(tmp[0])
            items = tmp[1:]
            self[nn] = SortedDict()
            for k, item in enumerate(['cent_gam/w_A', 'gam/(omegspi_max/4)', 'gamsq', 'del', 'gamr', 'gamr2', 'tmatasym']):
                self[nn][item] = float(items[k])


class OMFITeliteEigenfunction(OMFITascii, SortedDict):
    def __init__(self, filename, symmetric=False):
        OMFITobject.__init__(self, filename)
        SortedDict.__init__(self)
        self.symmetric = symmetric
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        lines = self.read().splitlines()
        self['date'] = lines[0]
        self['version'] = lines[1]
        self['eq_type'] = lines[2]
        self['n'] = int(lines[3])
        self['eq_data'] = np.array(list(map(float, (lines[4] + ' ' + lines[5]).split())))
        self['eq_par'] = np.array(list(map(int, (' '.join(lines[6:8])).split())))
        self['dummy'] = np.array(list(map(float, (' '.join(lines[8:12])).split())))
        tmp = np.array(list(map(float, (' '.join(lines[12:])).split())))
        tmp = np.reshape(tmp, (-1, self['eq_par'][9])).T
        self['xx'] = tmp[:, 0]
        self['psi'] = tmp[:, -1]
        if self.symmetric:
            self['eigh'] = tmp[:, 1:-1:1] + 1j * np.zeros(tmp[:, 1:-1:1].shape)
        else:
            self['eigh'] = tmp[:, 1:-2:2] + 1j * tmp[:, 2:-1:2]

    def peak(self, decimate=1):
        f = np.zeros(self['eigh'][::decimate].shape)
        for k1 in np.linspace(0, 2 * np.pi, 64)[:-1]:
            m1 = np.exp(1j * k1 * 2)
            f = np.max([f, np.real(self['eigh'][::decimate] * m1)], 0)
        return f

    def envelope(self, decimate=1):
        return np.max(self.peak(decimate), 1)

    def plot(self, fancy=False, decimate=None):
        from omfit_classes.utils_plot import plotc

        if decimate is None:
            decimate = int(np.ceil(1 + len(self['eigh']) / 10000.0))
        if fancy:
            tmp = self.peak(decimate)
            plotc(self['psi'][::decimate], tmp)
            pyplot.plot(self['psi'][::decimate], np.max(tmp, 1), color='k')
        else:
            y = self.envelope(decimate)
            x = self['psi'][::decimate]
            tmp = abs(np.trapz(y[np.where(x < 0.98)[0]], x[np.where(x < 0.98)[0]]))
            pyplot.plot(x, y / tmp)
        pyplot.ylim([1e-8, pyplot.ylim()[-1]])
        pyplot.xlabel('$\\psi$')
        pyplot.title('Eigenfunction n=%d' % self['eq_par'][5])


class OMFITelite2DEigenfunction(OMFITtree):
    def __init__(self, filename, psicontour=np.asarray([0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1])):
        self.dynaLoad = True
        self.filename = filename
        self.psicontour = psicontour

    @dynaLoad
    def load(self):
        self['xtopsi'] = OMFITelitextopsi(self.filename + '.xtopsi')
        self['fun2d'] = OMFITelitefun2d(self.filename + '.fun2d')
        self['xtopsi'].load()
        self['fun2d'].load()
        if len(self.psicontour) > 0:
            self['psiind'] = []
            for psi in self.psicontour:
                self['psiind'].append(np.argmin(abs(self['xtopsi']['psiinterp'] - psi)))
            self['rind'] = []
            for i in self['psiind']:
                self['rind'].append(
                    np.argmin(
                        abs(
                            np.asarray(self['fun2d']['xx'][1 : self['fun2d']['ns'] * self['fun2d']['nxplt'] : self['fun2d']['ns']])
                            - self['xtopsi']['rrmax'][i]
                        )
                    )
                )

    @dynaSave
    def save(self):
        self['xtopsi'].save()
        self['fun2d'].save()

    def plot(self):
        pyplot.tricontourf(self['fun2d']['xx'], self['fun2d']['zz'], self['fun2d']['f2d'], 30, cmap='rainbow')
        if len(self.psicontour) > 0:
            for i in self['rind']:
                pyplot.plot(
                    self['fun2d']['xx'][i * self['fun2d']['ns'] + 1 : (i + 1) * self['fun2d']['ns']],
                    self['fun2d']['zz'][i * self['fun2d']['ns'] + 1 : (i + 1) * self['fun2d']['ns']],
                    'k-',
                    linewidth=0.5,
                )
        pyplot.axis('equal')


class OMFITelitefun2d(OMFITascii, SortedDict):
    def __init__(self, filename):
        OMFITobject.__init__(self, filename)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        lines = self.read().splitlines()
        self['cdatetime'] = lines[0]
        self['codver'] = lines[1]
        self['shape'] = lines[2]
        self['npts'] = int(lines[3])
        set3 = np.array(list(map(float, (' '.join(lines[4:6])).split())))
        set2 = np.array(list(map(int, (' '.join(lines[6:8])).split())))
        self['nx'] = set2[9]
        self['nm'] = set2[4]
        set1 = np.array(list(map(float, (' '.join(lines[8:12])).split())))
        set0 = np.array(list(map(int, (lines[12]).split())))
        self['nxplt'] = set0[0]
        self['ns'] = set0[1]
        tmp = np.array(list(map(float, (' '.join(lines[13:])).split())))
        tmp = np.reshape(tmp, (-1, self['nxplt'] * self['ns'])).T
        self['xx'] = tmp[:, 0] / 100.0
        self['zz'] = tmp[:, 1] / 100.0
        self['bp'] = tmp[:, 2]
        self['f2d'] = tmp[:, 3]

    def plot(self):
        from matplotlib import pyplot

        pyplot.tricontourf(self['xx'], self['zz'], self['f2d'], 30, cmap='rainbow')
        pyplot.axis('equal')


class OMFITelitextopsi(OMFITascii, SortedDict):
    def __init__(self, filename):
        OMFITobject.__init__(self, filename)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        lines = self.read().splitlines()
        self['nxinterp'] = int(lines[0])
        tmp = np.array(list(map(float, (' '.join(lines[1:-1])).split())))
        tmp = np.reshape(tmp, (-1, self['nxinterp'])).T
        self['xinterp'] = tmp[:, 0]
        self['psiinterp'] = tmp[:, 1]
        self['rrmax'] = tmp[:, 2] / 100.0
        self['rmax'] = tmp[:, 3] / 100.0


class OMFITeliteAggregated(OMFITascii, SortedDict):
    def __init__(self, filename):
        OMFITascii.__init__(self, filename)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            tmp = f.readlines()
        dum = ' '.join([line.strip() for line in tmp])
        dum = list(map(float, dum.split()))
        try:
            dum = np.reshape(dum, (-1, 12))
        except Exception:
            dum = np.reshape(dum, (-1, 11))

        # in case the ELITE run did not finish
        self['alpha'] = dum[:, 0]
        self['pprime'] = dum[:, 1]
        self['jmax'] = dum[:, 2]
        self['jsep'] = dum[:, 3]
        self['gamma'] = dum[:, 4]
        self['ngam'] = dum[:, 5].astype('int')
        self['gamws'] = dum[:, 6]
        self['ngamws'] = dum[:, 7].astype('int')
        self['javg'] = dum[:, 8]
        self['gamws16'] = dum[:, 9]
        self['ngamws16'] = dum[:, 10].astype('int')
        if dum.shape[1] == 12:
            self['icount'] = dum[:, 11].astype('int')


class OMFITeqdat(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface to equilibria files generated by ELITE and BALOO (.dskbal files)

    NOTE: this object is "READ ONLY", meaning that the changes to the entries of this object will not be saved to a file. Method .save() could be written if becomes necessary

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename='', **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.readlines()

        for k, line in enumerate(lines):
            if 'npsi,nthet' in line:
                break
        if k == len(lines) - 1:
            raise ValueError('eqdat file should have a line with `npsi,nthet` in it')

        k += 1
        self['npsi'], self['npts'] = list(map(int, lines[k].split()))
        lsegment = int(np.ceil(self['npsi'] / 5.0))

        while k < len(lines) - 1:
            k += 1
            name = lines[k].strip()
            self[name] = []
            if name not in ['X', 'Z', 'poloidal field--bpol']:
                for seg in range(lsegment):
                    k += 1
                    self[name].extend(re.sub('D', 'E', lines[k]).split())
                self[name] = np.flipud(np.array(list(map(eval, self[name]))))
            else:
                for x in range(self['npts']):
                    self[name].append([])
                    for seg in range(lsegment):
                        k += 1
                        self[name][-1].extend(re.sub('D', 'E', lines[k]).split())
                    self[name][-1] = np.array(list(map(eval, self[name][-1])))
                self[name] = np.array(self[name])

        for name in self:
            if isinstance(self[name], np.ndarray) and len(self[name].shape) == 1:
                self[name] = self[name][::-1]

    @dynaLoad
    def plot(self, **kw):
        from omfit_classes.utils_plot import title_inside, autofmt_sharex

        items = [item for item in self if item not in ['npsi', 'npts', 'X', 'Z', 'poloidal field--bpol']]

        ax = None
        r = np.sqrt(len(items) + 1)
        c = np.ceil((len(items)) / np.sqrt(len(items))) + 1

        for k, item in enumerate(items):
            x = self.get('poloidal flux--psiv', np.arange(len(self[item])))
            x = (x - min(x)) / (max(x) - min(x))
            ax = pyplot.subplot(r, c, k + 1 + k // (c - 1) + 1)
            ax.plot(x, self[item], **kw)
            title_inside(item, y=0.8)
            ax.set_xlim([0, 1])

        nskip = self['npsi'] // 10
        ax0 = ax
        ax = pyplot.subplot(1, c, 1)
        ax.plot(self['X'][:, ::-nskip] / 100.0, self['Z'][:, ::-nskip] / 100.0, color=ax0.lines[-1].get_color())
        ax.set_aspect('equal')
        ax.set_frame_on(False)

        autofmt_sharex()


class OMFITbalstab(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface to `balstab` file from BALOO

    NOTE: this object is "READ ONLY", meaning that the changes to the entries of this object will not be saved to a file. Method .save() could be written if becomes necessary

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename='', **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            tmp = f.read().strip()
        tmp = re.sub('% ', '%_', tmp)
        tmp = re.sub(' s ', ' 0 ', tmp)
        tmp = re.sub(' u ', ' 1 ', tmp)
        tmp = tmp.split('\n')
        numbers = list(map(int, tmp[0].split()))
        headers = tmp[1].split()
        self.clear()
        for k in range(2, numbers[0] + 2):
            res = list(map(float, tmp[k].split()))
            res[0] = int(res[0])
            res[4] = int(res[4])
            self[res[0]] = {}
            for k1, h in enumerate(headers):
                if k1 == 0:
                    continue
                self[res[0]][h] = res[k1]


class OMFIToutbal(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface to `outbal` file from BALOO

    NOTE: this object is "READ ONLY", meaning that the changes to the entries of this object will not be saved to a file. Method .save() could be written if becomes necessary

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename='', **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        with open(self.filename, 'r') as f:
            lines = f.read().strip().split('\n')
        if 'PARZERO' in lines[0]:
            lines = lines[1:]  # skip loading this line if there is a parzero error
        tmp = []
        for k in range(int(len(lines) // 2)):
            tmp.append(lines[k * 2] + lines[k * 2 + 1])
        names = tmp[0].split()
        tmp = tmp[1:]
        for k, line in enumerate(tmp):
            tmp[k] = list(map(float, line.split()))
        tmp = np.array(tmp)
        for k, name in enumerate(names):
            self[name] = tmp[:, k]

    @dynaLoad
    def plot(self, **kw):
        from omfit_classes.utils_plot import title_inside

        items = [item for item in self if item not in ['surf', 'psi_norm']]

        r = np.sqrt(len(items) + 1)
        c = np.ceil((len(items)) / np.sqrt(len(items)))
        for k, item in enumerate(items):
            x = self.get('psi_norm', np.arange(len(self[item])))
            x = (x - min(x)) / (max(x) - min(x))
            ax = pyplot.subplot(r, c, k + 1)
            ax.plot(x, self[item], **kw)
            title_inside(item, y=0.8)
            ax.set_xlim([0, 1])



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise


from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes import zdata
import numpy as np

__all__ = ['OMFITfastran']


class OMFITfastran(SortedDict, OMFITascii):
    """FASTRAN data files"""

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self, sorted=True)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        self.clear()
        tmp = zdata.read_formatted(self.filename)
        self.update(tmp)
        for k in self:
            if len(self[k]) == 1:
                self[k] = self[k][0]
            else:
                self[k] = np.array(self[k])

    @dynaSave
    def save(self):
        tmp = {}
        tmp.update(self)
        for k in self:
            self[k] = np.atleast_1d(self[k]).tolist()
        zdata.from_dict(self, self.filename)



import sys
import socketserver as SocketServer
import socket
import threading
import time
import os
import numpy as np
import struct


def b2s(obj):
    if isinstance(obj, bytes):
        for error_handling in ['strict', 'replace', 'ignore']:
            try:
                return obj.decode("utf-8", errors=error_handling)
            except Exception:
                pass
        raise RuntimeError('A bytes object was passed to b2s, but not handled')
    return obj


default_serve_port = 8883

# =======================
# helper functions
# =======================
def send_data(sock, path, data):
    payload = '{path}&{shape}&[{data}]'.format(path=path, shape=data.shape, data=','.join(['%f' % x for x in data.flatten()]))
    return send_msg(sock, payload)


def send_ask_info(sock, path):
    payload = '{path}&(?,?)'.format(path=path)
    return send_msg(sock, payload)


def send_info(sock, path, x_names, y_names):
    payload = '{path}&{x_names}&{y_names}'.format(path=path, x_names=repr(list(map(str, x_names))), y_names=repr(list(map(str, y_names))))
    return send_msg(sock, payload)


def send_msg(sock, msg):
    # Prefix each message with a 4-byte length (network byte order)
    if sys.version_info < (3, 0):
        msg = struct.pack('=I', len(msg)) + msg
        sock.sendall(msg)
    else:
        msg = bytes(msg, 'utf-8')
        msg = struct.pack('=I', len(msg)) + msg
        sock.sendall(msg)


def parse_data(data):
    try:
        path, shape, data = b2s(data).split('&')
    except Exception:
        raise Exception('Malformed input data')
    return path, np.reshape(eval(data), eval(shape))


def parse_info(data):
    path, x_names, y_names = b2s(data).split('&')
    return path, eval(x_names), eval(y_names)


def recv_msg(sock):
    # Read message length and unpack it into an integer
    raw_msglen = recvall(sock, 4)
    if not raw_msglen:
        return None
    msglen = struct.unpack('=I', raw_msglen)[0]
    # Read the message data
    return b2s(recvall(sock, msglen))


def recvall(sock, n):
    # Helper function to recv n bytes or return None if EOF is hit
    data = b''
    while len(data) < n:
        packet = sock.recv(n - len(data))
        if not packet:
            return None
        data += packet
    return data


# =======================
# client
# =======================
class btf_connect(object):
    def __init__(self, path, host=None, port=None):
        if host is None:
            host = os.environ.get('BTF_HOST', 'gadb-harvest.duckdns.org')
        if port is None:
            port = int(os.environ.get('BTF_PORT', default_serve_port))
        self.host = host
        self.port = port
        self.path = path
        self.connection = None
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def __enter__(self):
        try:
            self.sock.connect((self.host, self.port))
        except Exception:
            print('HOST:%s PORT:%s' % (self.host, self.port))
        return self

    def __exit__(self, exec_type, exec_value, exec_tb):
        self.sock.close()

    def info(self):
        send_ask_info(self.sock, self.path)
        data = recv_msg(self.sock)
        if data is None:
            raise ValueError('Error receiving infos for model: %s' % self.path)
        path, self.x_names, self.y_names = parse_info(data)
        return self.x_names, self.y_names

    def run(self, input):
        if isinstance(input, dict):
            if not hasattr(self, 'x_names') or not hasattr(self, 'y_names'):
                with btf_connect(host=self.host, port=self.port, path=self.path) as btf:
                    self.x_names, self.y_names = btf.info()
            print(self.x_names)
            print(self.y_names)
            input = np.array([input[name] for name in self.x_names]).T
        send_data(self.sock, self.path, input)
        path, output = parse_data(recv_msg(self.sock))
        if isinstance(input, dict):
            output = {name: output[:, k] for k, name in enumerate(self.y_names)}
        return output


def activateNets(nets, dB):
    '''
    :param nets: dictionary with OMFITbrainfuse objects (or path where to load NNs from)

    :param dB: dictionary with entries to run on

    :return: tuple with (out,sut,targets,nets,out_)
    '''
    if isinstance(nets, basestring):
        nets = {k: OMFITpath(file) for k, file in enumerate(glob.glob(nets))}
    elif not isinstance(nets, (list, tuple)):
        nets = {0: nets}
    net = list(nets.values())[0]

    with btf_connect(path=net.filename) as btf:
        inputNames, outputNames = btf.info()
    targets = array([dB[k] for k in outputNames]).T

    out_ = empty((len(atleast_1d(list(dB.values())[0])), len(outputNames), len(nets)))
    for k, n in enumerate(nets):
        with btf_connect(path=net.filename) as btf:
            out_[:, :, k] = btf.run(dB)
    out = mean(out_, -1)
    sut = std(out_, -1)
    return out, sut, targets, nets, out_


# =======================
# server
# =======================
if __name__ == "__main__":
    import tensorflow as tf
    from tensorflow.python.platform import gfile

    __file__ = os.path.abspath(__file__)

    serve_port = default_serve_port
    if len(sys.argv) > 1:
        serve_port = int(sys.argv[1])

    models = {}

    def activate(path, input):
        '''
        high level function that handles models buffering
        :param path: path of the model
        :param input: input array (or None to get xnames,ynames)
        :return: output array or xnames,ynames
        '''
        if not path.startswith(os.sep):
            path = os.path.split(__file__)[0] + os.sep + path
        path = os.path.realpath(path)
        if path not in models:
            print('Loading %s' % path)
            models[path] = model(path=path)
        else:
            print('mtime %f' % os.path.getmtime(path))
            print('ltime %f' % models[path].load_time)
            if os.path.getmtime(path) > models[path].load_time:
                raise Exception('Model reload functionality not implemented')  # somehow the model does not get updated?
                # print('Updating %s'%path)
                # models[path].close()
                # models[path]=model(path=path)
        if input is None:
            return models[path].x_names, models[path].y_names
        else:
            return models[path].activate(input)

    class model(tf.Session):
        def __init__(self, target='', graph=None, config=None, path=None):
            tf.Session.__init__(self, target=target, graph=graph, config=config)
            self.__enter__()
            if not path.startswith(os.sep):
                path = os.path.split(__file__)[0] + os.sep + path
            path = os.path.realpath(path)
            self.path = path
            self.load_time = os.path.getmtime(self.path)
            with gfile.FastGFile(self.path, 'rb') as f:
                graph_def = tf.GraphDef()
                graph_def.ParseFromString(f.read())

            self.y, self.x_names, self.y_names = tf.import_graph_def(
                graph_def, return_elements=['unnormalized_nn/y:0', 'x_names:0', 'y_names:0'], name=''
            )
            self.x_names = self.x_names.eval()
            self.y_names = self.y_names.eval()

        def activate(self, input):
            '''
            :param input: input array
            :return: output array
            '''
            print('Serving %s' % self.path)
            return self.run(self.y, feed_dict={'x:0': input})

    class ThreadedTCPRequestHandler(SocketServer.BaseRequestHandler):
        def handle(self):
            msg = recv_msg(self.request)
            if msg is None:
                print("{}: {}".format(self.client_address[0], '-- no message --'))
                return
            print("{}: message length {}".format(self.client_address[0], len(msg)))
            query = msg.split('&')[1]
            # respond to info request
            if query == '(?,?)':
                print('INFO-MODE')
                path = msg.split('&')[0]
                x_names, y_names = activate(path=path, input=None)
                send_info(self.request, path, x_names, y_names)
            # respond to data request
            else:
                print('DATA-MODE: serve starts')
                while True:
                    try:
                        if msg is not None:
                            path, input = parse_data(msg)
                            output = activate(path=path, input=input)
                        send_data(self.request, path, output)
                        msg = recv_msg(self.request)
                    except Exception as _excp:
                        if (
                            'Broken pipe' in repr(_excp)
                            or 'Connection reset by peer' in repr(_excp)
                            or 'Protocol wrong type for socket' in repr(_excp)
                        ):
                            print('DATA-MODE: serve ends')
                            break
                        else:
                            raise

    class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
        pass

    server = ThreadedTCPServer(('0.0.0.0', serve_port), ThreadedTCPRequestHandler)
    print('-- BASH shell --')
    print("export BTF_HOST=%s" % socket.gethostname())
    print("export BTF_PORT=%d" % serve_port)
    print('-- TCSH shell --')
    print("setenv BTF_HOST %s" % socket.gethostname())
    print("setenv BTF_PORT %d" % serve_port)
    try:
        server.serve_forever()
    except (KeyboardInterrupt, SystemExit):
        pass
    finally:
        server.shutdown()
        server.server_close()



# load ODS from IMAS
def load():
    ods = load_omas_imas_remote(
        imas_settings['serverPicker'], user=None, machine=imas_settings['machine'], pulse=imas_settings['pulse'], run=imas_settings['run']
    )
    loc = parseLocation(location)
    eval(buildLocation(loc[:-1]))[loc[-1]] = ods
    if postcommand:
        postcommand()


# save ODS to IMAS
def save():
    save_omas_imas_remote(
        imas_settings['serverPicker'],
        eval(location),
        user=None,
        machine=imas_settings['machine'],
        pulse=imas_settings['pulse'],
        run=imas_settings['run'],
        new=True,
    )
    if postcommand:
        postcommand()


# delete ODS from tree
def delete():
    loc = parseLocation(location)
    del eval(buildLocation(loc[:-1]))[loc[-1]]


# =======================
defaultVars(
    base_root=None,
    action=['load', 'save', 'save_user', 'load_user', 'run'][0],
    location=None,
    postcommand=None,
    imas_settings_location=None,
    state='normal',
    updateGUI=False,
)

rl = relativeLocations(base_root)
root = rl['root']

# set ODS location based on action
if location is None:
    if action.startswith('load'):
        location = "root['INPUTS']['ods']"
    elif action.startswith('save'):
        location = "root['OUTPUTS']['ods']"

# if ODS already exists and action is load, allow de-loading
if location is not None:
    tmp = parseLocation(location)

if action == 'load' and tmp[-1] in eval(buildLocation(tmp[:-1])):
    OMFITx.Button('Load a different ODS', delete, updateGUI=True)
    OMFITx.End()

# other actions
else:
    # imas settings location
    if imas_settings_location is None:
        imas_settings_location = rl['thisName'] + "['__scratch__']"
    imas_settings = eval(imas_settings_location)

    # imas servers
    IMAS_SERVERS = {'ITER': 'iter_login', 'WPCD': 'itm_gateway', 'KFE': 'sophie', 'JET': 'heimdall', 'MAST': 'freia'}
    default_server = 'itm_gateway'
    if 'serverPicker' in root['SETTINGS']['REMOTE_SETUP'] and root['SETTINGS']['REMOTE_SETUP']['serverPicker'] in IMAS_SERVERS.values():
        default_server = root['SETTINGS']['REMOTE_SETUP']['serverPicker']

    # GUI entries
    with OMFITx.same_row():
        OMFITx.ComboBox(imas_settings_location + "['serverPicker']", IMAS_SERVERS, 'Server', default=default_server)
        OMFITx.Entry(imas_settings_location + "['machine']", 'Machine', default=root['SETTINGS']['EXPERIMENT']['device'])
    with OMFITx.same_row():
        OMFITx.Entry(imas_settings_location + "['pulse']", 'Pulse', check=is_int, default=root['SETTINGS']['EXPERIMENT']['shot'])
        if action == 'run':
            OMFITx.Entry(imas_settings_location + "['run_in']", 'Run in', check=is_int, default=root['SETTINGS']['EXPERIMENT']['time'])
            OMFITx.Entry(imas_settings_location + "['run_out']", 'Run out', check=is_int, default=root['SETTINGS']['EXPERIMENT']['time'])
        else:
            OMFITx.Entry(imas_settings_location + "['run']", 'Run', check=is_int, default=root['SETTINGS']['EXPERIMENT']['time'])

    # handle actions
    if action == 'run':
        pass
    elif action.startswith('load'):
        if action == 'load_user':
            OMFITx.Button('Load ODS from IMAS', postcommand, state=state, updateGUI=updateGUI)
        else:
            OMFITx.Button('Load ODS from IMAS', load, state=state, updateGUI=True)
    elif action.startswith('save'):
        if action == 'save_user':
            OMFITx.Button('Save ODS to IMAS', postcommand, state=state, updateGUI=updateGUI)
        else:
            OMFITx.Button('Save ODS to IMAS', save, state=state, updateGUI=updateGUI)
    else:
        raise ValueError('Action `%s` was not recognized' % action)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.utils_math import *
from omfit_classes.omfit_base import OMFITtree
from omfit_classes.omfit_osborne import OMFITpFile
from omfit_classes.omfit_eqdsk import OMFITgeqdsk, from_mds_plus
from omfit_classes.omfit_data import OMFITncDataset, OMFITncDynamicDataset, importDataset
from omas import ODS, omas_environment, cocos_transform, define_cocos
from omfit_classes.omfit_omas_utils import add_generic_OMFIT_info_to_ods
from omfit_classes.omfit_rdb import OMFITrdb
from omfit_classes.omfit_mds import OMFITmdsValue
from omfit_classes.omfit_profiles import OMFITprofiles

import inspect
import numpy as np
from uncertainties import unumpy, ufloat
from uncertainties.unumpy import std_devs, nominal_values
from scipy import constants

np.seterr(invalid='ignore', divide='ignore')


__all__ = ['available_iri_results', 'load_iri_results']


def available_iri_results(device='DIII-D', shot=-1, run_by=None, tag=None, settings=None, ignore_ignore=False):
    r"""
    Search for avaliable IRI runs by device and shot number. Optionally, also search with run_by and tag values.
    If multiple search criteria is given, then the search only returns records
    that satisfies all criteria. This function returns a dictionary containing
    all matching data that are

    :param device: string
        Name of device for which the analysis was done. Currently only DIII-D is supported.

    :param shot: int
        The shot number to search for.

    :param run_by: string [optional]
        The run_by username to search for in IRI records. Production runs are run by the user `d3dsf` which is the default

    :param tag: string [optional].
        The tag to search for in IRI records. There are currently three main tags: 'CAKE01' (no MSE), 'CAKE02'
        (with MSE) and 'CAKE_FDP'. 'CAKE01' and 'CAKE02' have 50 ms time resolution and 129x129 equilibria. 'CAKE_FDP'
        has 20 ms time resolution and 257x257 equilibria.

    :param settings: string [optional]
        The seetings string to search for in IRI records. For example '1FWDyrS` is defalut for knot optimized MSE
        constrained equilibria. See documentation for `code` in between_shot_autorun() in the CAKE module for full
        details.

    :param ignore_ignore: bool [optional]
        If this flag is set, then the 'ignore' field in the IRI metadata tables
        will be ignored. Thus this function will then return records that have
        been marked 'ignore'. Defaults to False.
    """

    if ignore_ignore:
        printw(
            """WARNING: You are ignoring the ignore flag. This will potentailly return results that are wrong,
            unphysical, or invalid! Use with EXTREAM CAUTION!"""
        )

    rdb = OMFITrdb(db='code_rundb', server='d3drdb')  # This where the iri metadata tables lives for now.
    select_string = f"SELECT * FROM iri_run_log WHERE experiment='{device}' AND shot={shot}"
    if shot != -1:
        select_string = select_string + f" AND shot={shot}"
    if run_by is not None:
        select_string = select_string + f" AND run_by='{run_by}'"
    if tag is not None:
        select_string = select_string + f" AND tag='{tag}'"
    if not ignore_ignore:
        select_string = select_string + f" AND ignore='False'"

    runs = rdb.select(select_string)
    n_runs = len(runs)
    if n_runs < 1:
        printw(f"No IRI runs found for {device} shot# {shot} run_by {run_by} with tag={tag}, settings={settings}.")
        return {}
    else:
        # we need to sort out the results in something easier to read and deal with.
        runs_dict = {}
        for ii in runs:
            run_dict = runs[ii]
            runid = run_dict['IRI_ID']
            run_dict[
                'description'
            ] = f"IRI_ID={runid} by {run_dict['RUN_BY']} with tag: {run_dict['TAG']} and comment: {run_dict['COMMENTS']}"

            # search for result uploads
            run_dict['results_uploaded'] = {}
            uploads = rdb.select(f"SELECT * FROM iri_upload_log WHERE iri_id={runid}")
            for jj in uploads:
                upload = uploads[jj]
                code = upload.pop('CODE_NAME', 'unknown code')
                run_dict['results_uploaded'][code] = upload

            runs_dict[runid] = run_dict

    return runs_dict


def load_iri_results(runs_dict):
    r"""
    Loads IRI results as described in a runs dictionary, and outputs to a OMFITtree()
    ex. OMFIT['iri_data'] = load_iri_results(runs_dict).

    :param runs_dict: dictionary
        Dictionary of metadata describing uploaded IRI data. It should come from
        the avaliable_iri_data() function. But editing and deleting of records
        is allowed as long as the hierarchy is preserved.
    """
    destination_tree = OMFITtree()
    for iri_id in runs_dict:
        destination_tree[iri_id] = OMFITtree()
        for code_upload in runs_dict[iri_id]['results_uploaded']:
            upload = runs_dict[iri_id]['results_uploaded'][code_upload]
            destination_tree[iri_id][code_upload] = load_by_code(
                code_upload, upload['UPLOAD_SERVER'], upload['UPLOAD_ID'], tree=upload['UPLOAD_TREE']
            )
    return destination_tree


def load_by_code(code, server, upload_id, tree=None):
    r"""
    Actual loading interface. It wraps underlying loading functions for various IRI related codes, and calls the
    appropriate one depending on which is needed. Returns OMFITtree() containing loaded results.

    :param runs_dict: 'string'
        The name of the code as recorded to the metadata. Currently supports 'OMFIT_CAKE_EFIT', and 'OMFIT_CAKE_PROF').

    :param server: 'string'
        The name of the sever where the data is stored. It should come from the metadata tables/record.

    :param upload_id: int
        The upload_id to load. Should come from the metadata record.
    """
    if code == 'OMFIT_CAKE_EFIT':
        eq_tree = OMFITtree()
        if tree is None:
            printd(f"No tree argument provided to load_by_code() for code = OMFIT_CAKE_EFIT. Tree assumed to be EFIT.")
            tree = 'EFIT'
        # from equilibriums, we need to find out the times that were uploaded.
        eq_times = get_EFIT_times_from_mds(server, tree, upload_id)
        eq_tree = from_mds_plus(device=server, shot=upload_id, times=eq_times, exact=True, snap_file=tree, get_afile=True, get_mfile=True)
        return eq_tree

    elif code == 'OMFIT_CAKE_PROF':
        prof = OMFITprofiles('')
        prof.from_mds(server, upload_id)
        return prof
    else:
        printw(f"OMFIT_IRI does not understand the code designation {code}. Results from that code is not loaded.")
        return OMFITtree()


def get_EFIT_times_from_mds(server, tree, upload_id):
    times_mds = OMFITmdsValue(server=server, treename=tree, shot=upload_id, TDI=f"dim_of(\{tree}::TOP.RESULTS.GEQDSK.CPASMA)")
    times = times_mds.data()
    return times



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise


import requests
import datetime
from urllib.parse import quote_plus
import numpy as np

__all__ = [
    'get_gh_remote_org_repo_branch',
    'on_latest_gh_commit',
    'convert_gh_time_str_datetime',
    'get_OMFIT_GitHub_token',
    'set_OMFIT_GitHub_token',
    'OMFITgithub_paged_fetcher',
    'get_pull_request_number',
    'get_git_branch_info',
    'post_comment_to_github',
    'find_gh_comments',
    'delete_matching_gh_comments',
    'edit_github_comment',
    'set_gh_status',
]


def get_git_branch_info(
    remote=None,
    org=None,
    destination_org=None,
    repository=None,
    branch=None,
    url=None,
    omfit_fallback=True,
    no_pr_lookup=False,
    return_pr_destination_org=True,
    server=None,
):
    """
    Looks up local name for upstream remote repo, GitHub org, repository name, current branch, & open pull request info
    All parameters are optional and should only be provided if trying to override some results

    :param remote: string [optional]
        Local name for the upstream remote.
        If None, attempts to lookup with method based on git rev-parse.

    :param org: string [optional]
        The organization that the repo is under, like 'gafusion'.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to gafusion on failure.

    :param destination_org: string [optional]
        Used for cross-fork pull requests: specify the destination org of the pull request.
        The pull request actually exists on this org, but it is not where the source branch lives.
        If None it defaults to same as org

    :param repository: string [optional]
        The name of the repo on GitHub.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to OMFIT-source on failure.

    :param branch: string [optional]
        Local/remote name for the current branch
        NOTE: there is an assumption that the local and upstream branches have same name

    :param url: string [optional]
        Provided mainly for testing.
        Overrides the url that would be returned by `git config --get remote.origin.url`.

    :param omfit_fallback: bool
        Default org and repository are gafusion and OMFIT-source instead of None and None in case of failed lookup.

    :param no_pr_lookup: bool
        Improve speed by skipping lookup of pull request number

    :param return_pr_destination_org: bool
        If an open pull request is found, changes remote, org, repository,
        and branch to match the destination side of the pull request.
        If there is no pull request or this flag is False,
        remote/org/repo/branch will correspond to the source.

    :param server: string [optional]
        The server of the remote - usually github.com, but could also be something like vali.gat.com.

    :return: tuple containing 4 strings and a dict, with elements to be replaced with None for lookup failure
        remote (str), org (str), repository (str), branch (str), pull_request_info (dict)
    """

    def _parse_remote_url(url):
        """
        A utility function to parse a git remote url

        :return: protocol, server, org, repo
        """
        if not url.endswith('.git'):
            url = url + '.git'
        # https://stackoverflow.com/a/4667014/6605826
        remote_re = r'(git@(.+?):|https://(.+?)/)(.+?)/(.+?)\.git$'
        m = re.search(remote_re, url)
        if m:
            server = m.group(2) if m.group(2) else m.group(3)
            return m.group(1), server, m.group(4), m.group(5)
        raise ValueError('Unable to parse remote url %s' % url)

    pr_info = None  # Default unless overwritten by something real

    if remote is not None and branch is not None:
        revparse = '{}/{}'.format(remote, branch)
    else:
        revparse = repo('rev-parse --abbrev-ref @{u}')

    # Jenkins server defines the GIT_BRANCH environment variable
    if len(revparse.split('/')) < 2:
        revparse = os.environ.get('GIT_BRANCH', revparse)
        # If this is a pr branch, then we can directly look up the relevant info (if requested)
        if '/pr/' in revparse and not no_pr_lookup and branch is None:
            dest_repo, _, number, __ = revparse.split('/')
            assert _ == 'pr', 'Upstream pr pattern %s not parsed correctly' % revparse
            assert __ == 'head', 'Upstream pr pattern %s not parsed correctly' % revparse
            url = repo('config --get remote.{}.url'.format(dest_repo))
            protocol_, server_, org_, repository_ = _parse_remote_url(url)
            pr_info = OMFITgithub_paged_fetcher(org=org_, path='pulls/%s' % number, repository=repository_).fetch()[0]
            if not return_pr_destination_org:
                printw('pr branch detected, but return_pr_destination_org is False')
                branch = branch or pr_info.get('head', {}).get('ref', None)
                full_name = pr_info.get('head', {}).get('repo', {}).get('full_name', "").split('/')
                org = org or (full_name[0] if len(full_name) is 2 else None)
                repository = repository or (full_name[1] if len(full_name) is 2 else None)
            else:
                remote = remote or dest_repo
                org = org or org_
                repository = repository or repository_
                branch = branch or pr_info.get('base', {}).get('ref', None)
            return remote, org, repository, branch, pr_info

    # Handle case of local branch with no upstream: we can't do much here
    if len(revparse.split('/')) < 2:
        printw('Warning: failed to determine remote/org/repo/branch info from git rev-parse: {}'.format(revparse))
        print('You probably are on a local branch that does not have an upstream.')
        branch = branch or repo.active_branch()[0]
        if omfit_fallback:
            if org is not None and repository is not None:
                print('org and repository were specified, so they are fine')
            else:
                print('org and repository will default to gafusion and OMFIT-source if not overridden')
            org = org or 'gafusion'
            repository = repository or 'OMFIT-source'
        printw('Returning: ({}, {}, {}, {}, {})'.format(remote, org, repository, branch, pr_info))
        return remote, org, repository, branch, pr_info

    # Get primary remote and upstream branch information
    remote_ = revparse.split('/')[0]
    branch_ = revparse.split('/')[-1]
    remote = remote or remote_
    branch = branch or branch_
    url = url or repo('config --get remote.{}.url'.format(remote))

    protocol_, server_, org_, repository_ = _parse_remote_url(url)
    org = org or org_
    repository = repository or repository_
    server = server or server_

    if no_pr_lookup:
        printd('Returning remote, org, repository, and branch before pull request lookup; pr_info will be None')
        return remote, org, repository, branch, pr_info

    # Build a list of remotes to check, starting with the first one
    destination_org = destination_org or org
    remotes = [remote]
    urls = [url]
    destination_orgs = [destination_org]
    repositories = [repository]
    servers = [server]
    for new_remote in repo('remote').split('\n'):
        if new_remote not in remotes:
            url = repo('config --get remote.{}.url'.format(new_remote))
            protocol, server, org_, repository_ = _parse_remote_url(url)
            remotes.append(new_remote)
            urls.append(url)
            destination_orgs.append(org_)
            repositories.append(repository_)
            servers.append(server)
            printd('    Found new remote to check: {} with org = {}, repo = {}'.format(new_remote, org, repository))
    # Make sure gafusion/OMFIT-source is in the list
    if 'gafusion' not in destination_orgs:
        destination_orgs.insert(1, 'gafusion')
        remotes.insert(1, None)
        repositories.insert(1, 'OMFIT-source')
        branches.insert(1, 'unstable')
        servers.insert(1, 'github.com')
    # Prioritize gafusion and check it first or second, but not after a bunch of other forks
    if 'gafusion' not in destination_orgs[0:2]:
        g_idx = destination_orgs.index('gafusion')
        for thing in [destination_orgs, remotes, repositories, servers, urls]:
            thing.insert(1, thing.pop(g_idx))

    # Loop through list of possible remote orgs and search for an open pull request
    head = '{org:}:{branch}'.format(org=org, branch=branch)  # The source branch, not the destination
    i = j = 0
    info = [{'empty': None}]
    printd('Searching for open pull requests for {}/{}:{} with destination_org in {}'.format(org, repository, branch, destination_orgs))
    while i < len(destination_orgs) and info[0].get('number', None) is None:
        if servers[i] != 'github.com':
            i += 1
            continue
        a = OMFITgithub_paged_fetcher(org=destination_orgs[i], path='pulls', selection=dict(head=head), repository=repository)
        j = i
        info = a.fetch()
        if not len(info):
            info = [{}]
        if not isinstance(info[0], dict):
            info = [{}]
        printd(
            "    PR lookup attempt i = {i:}, destination_org = {do:}, info[0].get('number', None) = {num:}".format(
                i=i, do=destination_orgs[i], num=info[0].get('number', None)
            )
        )
        info[0].setdefault('org', destination_orgs[i])
        i += 1

    if info[0].get('number', None) is None:
        printd('This branch ({}:{}) does not seem to have an open pull request.'.format(org, branch))
    else:
        pr_info = info[0]
        if return_pr_destination_org:
            printd(
                'Open pull request {}#{} found! Reassigning remote, org, repository, and branch to correspond to '
                'the destination side of the pull request.'.format(destination_orgs[j], pr_info.get('number', None))
            )
            remote = remotes[j]
            org = destination_orgs[j]
            repository = repositories[j]
            branch = pr_info.get('base', {}).get('ref', None)
            printd('    remote = {remote:}, org = {org:}, repository = {repository:}, branch = {branch:}'.format(**locals()))

    return remote, org, repository, branch, pr_info


def on_latest_gh_commit():
    """
    Returns true if the current working commit is the same as the github latest commit for this branch.
    """
    remote, org, repository, branch, pr_info = get_git_branch_info(return_pr_destination_org=False)
    if org and repo and branch:
        headers = {'Authorization': 'token ' + get_OMFIT_GitHub_token()}
        url = "https://api.github.com/repos/{}/{}/commits?sha={}".format(org, repository, quote_plus(branch))
        print("URL : " + url)
        response = requests.get(url, headers=headers)
        local_hash = repo.get_hash('HEAD~0')
        if response.status_code == 200:
            commits = response.json()
            if commits and commits[0].get('sha', local_hash) != local_hash:
                return False

    else:
        print("Could not find corresponding GitHub branch. Assuming current commit is latest.")
    return True


def get_gh_remote_org_repo_branch(**kw):
    """
    Looks up local name for upstream remote repo, GitHub org, repository name, and current branch
    Like calling get_git_branch_info with `no_pr_lookup=True`

    :return: tuple containing 4 strings
        remote, org, repository, branch
    """
    kw['no_pr_lookup'] = True  # Make sure no one can override this; the function wouldn't make any sense otherwise.
    return get_git_branch_info(**kw)[0:4]


def _get_OMFIT_GitHub_credential():
    user = OMFIT['MainSettings']['SERVER'].get('GITHUB_username', '')
    if not len(user):
        raise ValueError('No GitHub username has been set!')
    return user + '@token.github.com:0'


def get_OMFIT_GitHub_token(token=None):
    """
    :param token: string or None
        Token for accessing GitHub
        None triggers attempt to decrypt from $GHUSER@token.github.com credential file
        Must be set up in advance with set_OMFIT_GitHub_token() function

    :return: GitHub token
    """
    if not token:
        _, token = decrypt_credential(_get_OMFIT_GitHub_credential())
    if len(token) == 40:
        return token
    if not len(token):
        raise ValueError(
            'See https://github.com/settings/tokens/new to create a token, ' 'then use set_OMFIT_GitHub_token() to encrypt and store it.'
        )
    raise ValueError('string `%s` is not a valid token' % token)


def set_OMFIT_GitHub_token(token):
    """
    :param token: 40 chars Token string to be encrypted in $GHUSER@token.github.com credential file
    """
    cred = _get_OMFIT_GitHub_credential()
    if not len(token):
        return reset_credential(credential=cred)
    if len(token) != 40:
        raise ValueError('string `%s` is not a valid token' % token)
    try:
        encrypt_credential(credential=cred, password='', otp=token)
        printi('GitHub token stored in:', cred)
    except Exception as e:
        printe('GitHub token save failed:', repr(e))


def convert_gh_time_str_datetime(t):
    """
    Convert a GitHub (gh) time string to a datetime object

    :param t: A time string like
    """
    if t is None:
        return None
    y, M, d = list(map(int, t.split('T')[0].split('-')))
    h, m, s = list(map(int, t.split('T')[1].strip('Z').split(':')))
    return datetime.datetime(y, M, d, h, m, s)


class OMFITgithub_paged_fetcher(list):
    """
    Interact with GitHub via the GitHub api: https://developer.github.com/v3/
    to fetch data from a path
    https://api.github.com/repos/{org}/{repo}/{path}
    that has paged results
    """

    def __init__(self, org=None, repository=None, path='comments', token=None, selection=None):
        """
        https://api.github.com/repos/{org}/{repo}/{path}

        :param org: string [optional] The organization that the repo is under, like 'gafusion'.
            If None, attempts to lookup with method based on git rev-parse.
            Falls back to gafusion on failure.

        :param repository: string [optional]
            The name of the repo on GitHub.
            If None, attempts to lookup with method based on git rev-parse.
            Falls back to OMFIT-source on failure.

        :param path: string
            The part of the repo api to access

        :param token: string or None
            Token for accessing GitHub
            None triggers attempt to decrypt from file (must be set up in advance).
            Passed to get_OMFIT_GitHub_token.

        :param selection: dict
            A dictionary such as {'state':'all'}
        """
        self.token = get_OMFIT_GitHub_token(token)
        self.path = path
        # Avoid infinite recursion
        if org is None or repository is None:
            remote, org, repository, branch = get_gh_remote_org_repo_branch(org=org, repository=repository)
        self.url = 'https://api.github.com/repos/{org}/{repository}/{path}'.format(**locals())
        if selection is None:
            self.selection = {}
        else:
            self.selection = selection
        self.current_page = 1
        self.find_last_page()

    def find_last_page(self):
        """
        Find the last page number and sets self.last_page

        :returns: self.last_page
        """
        req = requests.get(self.url + self.get_sel_str(page=1), headers={'Authorization': 'token %s' % self.token})
        if hasattr(req, 'links') and 'last' in req.links and 'url' in req.links['last']:
            last = self.last_page_url = req.links['last']['url']
            for s in last.split('?')[1].split('&'):
                if 'page' in s:
                    self.last_page = int(s.split('=')[1].strip())
                    break
        else:
            self.last_page = 1
            printd('Unable to find last page of GitHub paged results')
        return self.last_page

    def fetch(self):
        """
        Fetch the paged results from GitHub and store them in self

        :returns: self
        """
        if (self.current_page == self.last_page) and (self.last_page > 1):
            printd('Already fetched last page; not fetching again')
            return
        start = self.current_page
        for i in range(self.current_page, self.last_page + 1):
            ascii_progress_bar(i, start, self.last_page, mess='Getting GitHub %s pages' % self.path, newline=False)
            try:
                req = requests.get(self.url + self.get_sel_str(page=i), headers={'Authorization': 'token %s' % self.token})
            except KeyboardInterrupt:
                break
            json = req.json()
            # Some paths (such as requesting only a single pull request) only return a single dictionary
            if isinstance(json, dict):
                json = [json]
            self.extend(json)
            self.current_page = i
        return self

    def get_sel_str(self, **kw):
        r"""
        Figure out how to get the selection part of the url, such as ?state=all&page=1, etc.

        :param \**kw: Any variables (such as page) to override self.selection set at initialization

        :return: An ampersand '&' separated string of key=value (and the question mark)
        """

        sel = copy.deepcopy(self.selection)
        sel.update(kw)
        if len(sel):
            return '?' + '&'.join(['%s=%s' % (x[0], x[1]) for x in list(sel.items())])
        else:
            return ''

    @property
    def results(self):
        return self


def get_pull_request_number(return_info=False, **kw):
    """
    Gets the pull request number associated for the current git branch if there is an open pull request.

    Passes parameters org, destination_org, branch, and repository to get_git_branch_info().

    :param return_info: bool [optional]
        Return a dictionary of information instead of just the pull request number

    :return: int, dict-like, or None
        Pull request number if one can be found, otherwise None.
        If return_info: dictionary returned by OMFITgithub_paged_fetcher with 'org' key added. Contains 'number', too.
    """
    kw['no_pr_lookup'] = False  # Make sure no one can override this; the function wouldn't make any sense otherwise.
    pr_info = get_git_branch_info(**kw)[4]
    if return_info or pr_info is None:
        return pr_info
    else:
        return pr_info.get('number', None)


def find_gh_comments(thread=None, contains='automatic_regression_test_post', user=False, id_only=True, org=None, repository=None, **kw):
    r"""
    Looks up comments on a GitHub issue or pull request and searches for ones with body text matching `contains`

    :param thread: int or None
        int: issue or pull request number
        None: look up pull request number based on active branch name. Only works if a pull request is open.

    :param contains: string or list of strings
        Check for these strings within comment body text. They all must be present.

    :param user: bool or string [optional]
        True: only consider comments made with the current username (looks up GITHUB_username from MainSettings)
        string: only comments made by the specified username.

    :param id_only: bool
        Return only the comment ID numbers instead of full dictionary of comment info

    :param org: string [optional] The organization that the repo is under, like 'gafusion'.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to gafusion on failure.

    :param repository: string [optional]
        The name of the repo on GitHub. If None, attempts to lookup with
        method based on git rev-parse. Falls back to OMFIT-source on failure.

    :param \**kw: keywords to pass to OMFITgithub_paged_fetcher

    :return: list of dicts (id_only=False) or list of ints (id_only=True)
    """
    assert contains is not None, 'Must specify text to match.'
    _, org, repository, _ = get_gh_remote_org_repo_branch(org=org, repository=repository)
    if thread is None:
        thread = get_pull_request_number(org=org, repository=repository)
        if thread is None:
            print(
                'Automatic lookup of pull request # failed. Must provide a pull request or issue number to post '
                'comments to GitHub. Nothing will be posted. Goodbye.'
            )
    info = OMFITgithub_paged_fetcher(org=org, path='issues/{}/comments'.format(thread), repository=repository, **kw).fetch()
    matching_comments = [cmt for cmt in info if np.all([c in cmt.get('body', '') for c in tolist(contains)])]
    if user is True:
        user = OMFIT['MainSettings']['SERVER'].get('GITHUB_username', False)
    if user:
        matching_comments = [cmt for cmt in matching_comments if cmt.get('user', {}).get('login', None) == user]
        userm = user
    else:
        userm = '(any user)'
    ids = [comment.get('id', None) for comment in matching_comments]
    printd(
        'Found {} comments in {}/{}#{} containing {} and user matching {} with IDs: {}'.format(
            len(matching_comments), org, repository, thread, ' and '.join(tolist([repr(c) for c in contains])), userm, ids
        ),
        topic='omfit_github',
    )
    if id_only:
        return ids
    return matching_comments


def delete_matching_gh_comments(
    thread=None,
    keyword=None,
    test=True,
    token=None,
    org=None,
    repository=None,
    quiet=False,
    exclude=None,
    exclude_contain=None,
    match_username=None,
    **kw,
):
    r"""
    Deletes GitHub comments that contain a keyword. Use CAREFULLY for clearing obsolete automatic test report posts.

    :param thread: int [optional]
        Supply issue or comment number or leave as None to look up an open pull request # for the active branch

    :param keyword: string or list of strings
        CAREFUL! All comments which match this string will be deleted.
        If a list is provided, every substring in the list must be present in a comment.

    :param test: bool
        Report which comments would be deleted without actually deleting them.

    :param token: string or None
        Token for accessing GitHub
        None triggers attempt to decrypt from $GHUSER@token.github.com credential file
        Must be set up in advance with set_OMFIT_GitHub_token() function

    :param org: string [optional] The organization that the repo is under, like 'gafusion'.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to gafusion on failure.

    :param repository: string [optional]
        The name of the repo on GitHub.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to OMFIT-source on failure.

    :param quiet: bool
        Suppress print output

    :param exclude: list of strings [optional]
        List of CIDs to exclude / protect from deletion. In addition to actual CIDs, the special value of 'latest' is
        accepted and will trigger lookup of the matching comment with the most recent timestamp. Its CID will replace
        'latest' in the list.

    :param exclude_contain: list of strings [optional]
        If provided, comments must contain all of the strings listed in their body in order to qualify for exclusion.

    :param match_username: bool or string [optional]
        True: Only delete comments that match the current username.
        string: Only delete comments that match the specified username.

    :param \**kw: keywords to pass to find_gh_comments()

    :return: list of responses from requests (test=False) or list of dicts with comment info (test=True)
        response instances should have a `status_code` attribute, which is normally int(201) for successful
        GitHub posts and probably 4xx for failures.
    """

    printd(
        'delete_matching_gh_comments: thread = {thread:}, keyword = {keyword:}, test = {test:}, org = {org:}, '
        'repository = {repository:}, exclude = {exclude:}, exclude_contain = {exclude_contain:}, '
        'match_username = {match_username:}'.format(**locals()),
        topic='delete_comments',
    )

    def printq(*args):
        if not quiet:
            print(*args)
        return

    gafusion_fallback_allowed = org is None  # Don't allow fallback to gafusion if the original org is explicit

    # Setup github interfacing information
    token = get_OMFIT_GitHub_token(token=token)
    hed = {'Authorization': 'token ' + token}
    responses = []
    remote, org, repository, branch = get_gh_remote_org_repo_branch(org=org, repository=repository)

    thread = thread or get_pull_request_number(org=org, repository=repository)

    # Lookup comments
    info = find_gh_comments(
        thread=thread, contains=keyword, id_only=False, token=token, org=org, repository=repository, user=match_username, **kw
    )
    cids = [c['id'] for c in info]

    if (len(cids) == 0) and (org != 'gafusion') and gafusion_fallback_allowed:
        # This may be a cross-fork pull request where "org" is not gafusion but the pull request is on gafusion.
        # See if we can find the comments if we change the org to gafusion and search there.
        gafusion_info = find_gh_comments(
            thread=thread, contains=keyword, id_only=False, token=token, org='gafusion', repository=repository, user=match_username, **kw
        )
        gafusion_cids = [c['id'] for c in gafusion_info]
        if len(gafusion_cids) > 0:
            info = gafusion_info
            cids = gafusion_cids
            printd('Found {} comments by changing org to gafusion'.format(len(cids)), topic='delete_comments')
            org = 'gafusion'
        else:
            printd('Changing org to gafusion did not help find comments', topic='delete_comments')
    elif len(cids) == 0:
        printd(
            "Nothing found in search for comments to delete; fallback method would've been nice, but wasn't attempted",
            topic='delete_comments',
        )
        if org == 'gafusion':
            printd('Fallback to searching for comments on gafusion not attempted because org was already gafusion', topic='delete_comments')
        if not gafusion_fallback_allowed:
            printd(
                'Fallback to searching for comments on gafusion not allowed because '
                'org was explicit in call to delete_matching_gh_comments()',
                topic='delete_comments',
            )
    else:
        printd('Found some comments on the first attempt with org={}; skipping fallback search plan'.format(org), topic='delete_comments')

    # Handle exclusions
    exclude = tolist(exclude)
    if 'latest' in exclude and len(info):
        printq('Excluding latest matching comment...')
        dates = [datetime.datetime.strptime(c['created_at'], "%Y-%m-%dT%H:%M:%SZ") for c in info]
        exclude_ = cids[np.array(dates).argmax()]
        printq('  CID of latest comment is {}'.format(exclude_))
        for i in range(len(exclude)):
            if exclude[i] == 'latest':
                exclude[i] = exclude_
        if exclude_ not in exclude:
            printw('Warning: had to use fallback to  put {} in exclude list'.format(exclude_))
            exclude += [exclude_]
    else:
        printd('No exclusion for latest comment', topic='delete_comments')
    if exclude_contain is not None:
        for i in range(len(exclude)):
            for k in range(len(info)):
                if info[k]['id'] == exclude[i]:
                    for ec in tolist(exclude_contain):
                        if ec not in info[k]['body']:
                            printq(
                                'Revoking protection for {} (it is no longer excluded from deletion) because '
                                'it does not contain all of the substrings: {}'.format(exclude[i], exclude_contain)
                            )
                            exclude[i] = None
    else:
        printd('No special requirements for excluded comments', topic='delete_comments')

    info = [info_ for info_ in info if info_['id'] not in exclude]
    cids = [c['id'] for c in info]

    # Test mode: return info without trying to delete anything
    if test:
        printq(
            'Bad authorization; nothing would be deleted because of bad token.'
            if token is None
            else 'Authorization may be okay. You could actually delete these comments:'
        )
        printq('{} comments would be deleted in #{}.'.format(len(cids), thread))
        printq('No comments have actually been deleted because test=True')
        return info
    for cid in cids:
        url = 'https://api.github.com/repos/{org:}/{repo:}/issues/comments/{cid:}'.format(org=org, repo=repository, cid=cid)
        responses += [requests.delete(url, headers=hed)]
        printd('responses to deletion requests: {}'.format(responses), topic='delete_comments')
    printq('Ordered deletion of {} comments in {}/{}#{}'.format(len(responses), org, repository, thread))
    return responses


def _define_new_content(old_content, new_content, mode='replace_between', separator='---', close_separator=None):
    """
    Defines new content of a comment after editing.

    Used for assigning output of the edit_github_comment function and also predicting results.

    :param old_content: str
        Content of the comment's body before editing

    :param new_content: str
        New content to add (maybe as a replacement of some old content)

    :param mode: str
        Edit mode: 'replace', 'append', 'replace_between' separators

    :param separator: str
        Separator between old_content and new content

    :param close_separator: str [optional]
         A different separator at the end of the block of new content

    :return: str, None, or ValueError
        str: try to do the edit, using this string as the new comment body
        None: abort
        ValueError: This error should be raised. It's returned instead of raised here to help predict output in tests.
    """
    if new_content is None:
        if mode == 'replace_between':
            if separator in old_content:
                before = old_content.split(separator)[0]
                after = separator.join(old_content.split(separator)[1:])
                if close_separator is not None:
                    if close_separator in old_content:
                        after = close_separator.join(after.split(close_separator)[1:])
                    else:
                        after = ''
                elif separator in after:
                    after = separator.join(after.split(separator)[1:])
                else:
                    after = ''
                content = before + after
            else:
                print('No separators found in output. Aborting deletion of material between separators ' '({}).'.format(separator))
                return None
        else:
            return ValueError("Can't accept None as content except in mode 'replace_between'")
        # new_content == '' is acceptable for modes replace and append, but has special behavior for replace_between
    elif mode == 'replace':
        content = new_content
    elif mode == 'append' and separator is None:
        content = old_content + '\n' + new_content
    elif mode == 'append':
        content = old_content + separator + new_content
        if close_separator is not None:
            content += close_separator
    elif mode == 'replace_between' and separator is None:
        return ValueError('Using mode: replace between separators requires specifying a separator.')
    elif mode == 'replace_between':
        if separator in old_content:
            if close_separator is None:
                content = old_content.split(separator)
                content[1] = new_content
                content = separator.join(content)
            else:
                before = old_content.split(separator)[0]
                after = separator.join(old_content.split(separator)[1:])
                after = close_separator.join(after.split(close_separator)[1:])
                content = before + separator + new_content + close_separator + after
        else:
            content = old_content + separator + new_content
            if close_separator is not None:
                content += close_separator
    else:
        return ValueError('Unrecognized mode: {}'.format(mode))
    return content


def edit_github_comment(
    comment_mark=None, separator='---', new_content=None, mode='replace_between', thread=None, org=None, repository=None, token=None
):
    """
    Edits GitHub comments to update automatically generated information, such as regression test reports.

    :param comment_mark: str or None
        None: edit top comment.
        str: Search for a comment (not including top comment) containing comment_mark as a substring.

    :param new_content: str
        New content to put into the comment

        Special cases:
        If content is None and mode == 'replace_between':
            Separate close separator and close separator present in target: del between 1st open sep and next close sep
            Separate close separator & close separator not present in target: del everything after 1st open sep
            All same separator & one instance present in target: delete it and everything after
            All same separator & multiple instances present: delete the first two and everything in between.
        If content is None and mode != 'replace_between', raise ValueError
        If None and mode == 'replace_between', but separator not in comment, abort but do not raise.

    :param separator: str or list of strings
        Substring that separates parts that should be edited from parts that should not be edited.
        `'---'` will put a horizontal rule in GitHub comments, which seems like a good idea for this application.
        If this is a list, the first and second elements will be used as the opening and closing separators to allow for
        different open/close around a section.

    :param mode: str
        Replacement behavior.
        'replace': Replace entire comment. Ignores separator.
        'append': Append new content to old comment. Places separator between old and new if separator is supplied.
           Closes with another separator if separate open/close separators are supplied; otherwise just places one
           separator between the original content and the addition.
        'replace_between: Replace between first & second appearances of separator (or between open & close separators).
           Raises ValueError if separator is not specified.
           Acts like mode == 'append' if separator (or opening separator) is not already present in target comment.
        other: raises ValueError

    :param thread: int [optional]
        Issue or pull request number. Looked up automatically if not provided

    :param org: str [optional]
        GitHub org. Determined automatically if not supplied.

    :param repository: str [optional]
        GitHub repository. Determined automatically if not supplied.

    :param token: str
        Token for accessing GitHub. Decrypted automatically if not supplied.

    :return: response instance or None
        None if aborted before attempting to post, otherwise response instance, which is an object generated
        by the requests module. It should have a `status_code` attribute which is 2** on success
        and often 4** for failures.
    """
    if separator is not None:
        seps = tolist(separator)
        separator = seps[0]
        if len(seps) > 1:
            close_separator = seps[1]
        else:
            close_separator = None
    else:
        close_separator = None

    if separator is not None and '\n' not in separator:
        printw('edit_github_comment: Your life would be better if you included a line break in your separator.')

    # Figure out remote, org, etc. so ensure it's consistent throughout all function calls
    token = get_OMFIT_GitHub_token(token=token)
    remote, org_, repository_, branch, pr_info = get_git_branch_info(
        destination_org=org, repository=repository, no_pr_lookup=False, return_pr_destination_org=True
    )
    org = org or org_
    repository = repository or repository_
    thread = thread or (pr_info.get('number', None) if pr_info is not None else None)
    if thread is None:
        print(
            'Automatic lookup of pull request # failed. Must provide a pull request or issue number to post '
            'comments to GitHub. Nothing will be posted. Goodbye.'
        )
        return None

    # Get old content and cid if needed
    if comment_mark is not None:
        matching = find_gh_comments(contains=comment_mark, org=org, repository=repository, thread=thread, token=token, id_only=False)
        cid = matching[0]['id']
        old_content = matching[0]['body']
    elif mode in ['append', 'replace_between']:
        # The top "comment" isn't really a comment by GitHub's naming convention, apparently.
        old_content = OMFITgithub_paged_fetcher(org=org, repository=repository, path='issues/{}'.format(thread), token=token).fetch()[0][
            'body'
        ]
        cid = None
    else:
        old_content = ''  # Mode 'replace' just replaces, so no one cares what this is and we can save time fetching it.
        cid = None

    # Deal with different line break conventions; \r\n moves to the start of the line and then breaks, \n just breaks.
    # On Unix, \n should be used. \r\n is used by Windows and apparently web stuff like GitHub. It's only important
    # here because it breaks the str comparison. GitHub will convert \n into whatever it needs.
    if old_content is None:
        old_content = ''
    old_content = old_content.replace('\r\n', '\n')

    # Define content to post
    content = _define_new_content(old_content, new_content, mode=mode, separator=separator, close_separator=close_separator)
    if content is None:
        return None
    elif isinstance(content, ValueError):
        raise content

    # Edit the comment on GitHub!
    hed = {'Authorization': 'token ' + token}
    data = {'body': content}
    if comment_mark is None:
        # The issue itself (not a real comment)
        url = 'https://api.github.com/repos/{org:}/{repo:}/issues/{thread:}'.format(org=org, thread=thread, repo=repository)
    else:
        # A comment on the issue
        url = 'https://api.github.com/repos/{org:}/{repo:}/issues/comments/{cid:}'.format(org=org, repo=repository, cid=cid)
    response = requests.patch(url, json=data, headers=hed)

    return response


def post_comment_to_github(thread=None, comment=None, org=None, fork=None, repository=None, token=None):
    """
    Posts a comment to a thread (issue or pull request) on GitHub.
    Requires setup of a GitHub token to work.

    This function may be tested on fork='gafusion', thread=3208 if needed.

    Setup::

        1. Create a GitHub token with the "repo" (Full control of private repositories) box checked.
           See https://github.com/settings/tokens .

        2. [Optional] Safely store the token to disk by executing:
              set_OMFIT_GitHub_token(token)
           This step allows you to skip passing the token to the function every time.

    :param thread: int [optional]
        The number of the issue or pull request within the fork of interest
        If not supplied, the current branch name will be used to search for open pull requests on GitHub.

    :param comment: string
        The comment to be posted

    :param org: string [optional]
        Leave this as gafusion to post to the main OMFIT repo.
        Enter something else to post on a fork.

    :param fork: string [optional]
        Redundant with org. Use org instead. Fork is provided for backwards compatibility

    :param repository: string [optional]
        The name of the repo on GitHub.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to OMFIT-source on failure.
        You should probably leave this as None unless you're doing some testing,
        in which case you may use the regression_notifications repository under gafusion.

    :param token: string or None
        Token for accessing GitHub
        None triggers attempt to decrypt from $GHUSER@token.github.com credential file
        Must be set up in advance with set_OMFIT_GitHub_token() function

    :return: response instance
        As generated by requests.
        It should have a `status_code` attribute, which is normally int(201) for successful
        GitHub posts and probably 4xx for failures.
    """
    import requests

    if org is None:
        org = fork

    if comment is None:
        print('Provide a comment if you want to post a comment.')
        return

    remote, org_, repository_, branch, pr_info = get_git_branch_info(
        destination_org=org, repository=repository, no_pr_lookup=False, return_pr_destination_org=True
    )
    org = org or org_
    repository = repository or repository_
    thread = thread or (pr_info.get('number', None) if pr_info is not None else None)
    if thread is None:
        print(
            'Automatic lookup of pull request # failed. Must provide a pull request or issue number to post '
            'comments to GitHub. Nothing will be posted. Goodbye.'
        )
        return

    token = get_OMFIT_GitHub_token(token=token)
    hed = {'Authorization': 'token ' + token}
    data = {'body': comment}

    url = 'https://api.github.com/repos/{org:}/{repo:}/issues/{thread:}/comments'.format(org=org, thread=thread, repo=repository)
    response = requests.post(url, json=data, headers=hed)
    return response


def set_gh_status(
    org=None, repository=None, commit=None, state=None, context='regression/auto', description='', target_url=None, destination_org=None
):
    """
    Updates the status of a pull request on GitHub. Appears as green check mark or red X at the end of the thread.

    :param org: string [optional] The organization that the repo is under, like 'gafusion'.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to gafusion on failure.

    :param destination_org: string [optional]
        Used for cross-fork pull requests: specify the destination org of the pull request.
        The pull request actually exists on this org, but it is not where the source branch lives.
        Passed to get_pull_request_number when determining whether a pull request is open.
        Defines first org to check.
        If None it defaults to same as org

    :param repository: string [optional]
        The name of the repo on GitHub.
        If None, attempts to lookup with method based on git rev-parse.
        Falls back to OMFIT-source on failure.

    :param commit: commit hash or keyword
        'latest' or 'HEAD~0':
                Look up latest commit. This is appropriate if you have reloaded modules and classes as needed.
        'omfit' or None:
                use commit that was active when OMFIT launched.
                Appropriate if testing classes as loaded during launch and not reloaded since.
        else: treats input as a commit hash and will fail if it is not one.

    :param state: string or bool
        'success' or True: success -> green check
        'failure' or False: problem -> red X
        'pending' -> yellow circle

    :param context: string
        Match the context later to update the status

    :param description: string
        A string with a description of the status. Up to 140 characters. Long strings will be truncated.
        Line breaks, quotes, parentheses, etc. are allowed.

    :param target_url: string
       URL to associate with the status

    :return: response instance s generated by `requests`.
        It should have a `status_code` attribute, which is normally int(201) for successful GitHub posts and probably 4xx for failures.
    """
    if commit in ['latest', 'HEAD~0']:
        printd('Using latest commit (assuming modules and classes have been reloaded as needed)')
        sha = repo.get_hash('HEAD~0')
    elif commit in ['omfit', None]:
        printd('Using active commit at time of OMFIT launch')
        try:
            sha = repo.get_hash(repo_active_branch_commit)
        except NameError:
            repo_active_branch, repo_active_branch_commit, repo_str = repo.active_branch()
            sha = repo.get_hash(repo_active_branch_commit)
    else:
        sha = commit
    remote, org, repository, branch, pr_info = get_git_branch_info(org=org, repository=repository, destination_org=destination_org)
    url = 'https://api.github.com/repos/{org}/{repository}/statuses/{sha}'
    url = url.format(org=org, repository=repository, sha=sha)
    h = {'Authorization': 'token ' + get_OMFIT_GitHub_token()}
    final_state = 'success' if state is True else ('failure' if state is False else state)
    printd('set_gh_status(): state = {}, context = {}, description = {}, sha = {}'.format(final_state, context, description, sha))
    if target_url is None:
        target_url = os.environ.get('BUILD_URL', None)
    data = {
        'state': final_state,
        'context': context,
        'description': description[:140],  # GitHub status has a 140 char length limit
        'target_url': target_url,
    }
    response = requests.post(url, json=data, headers=h)
    return response


############################################
if '__main__' == __name__:
    test_classes_main_header()

    if False:  # remove test since token is encrypted with ssh key which has passphrase#os.environ.get('USER','') in ['eldond','smithsp']:
        comments = OMFITgithub_paged_fetcher()
        assert comments.last_page > 1
        assert convert_gh_time_str_datetime('2018-12-05T15:40:30Z') == datetime.datetime(2018, 12, 5, 15, 40, 30)
        post_comment_to_github(thread=3, comment='Command line import test comment', fork='smithsp')
    else:
        print('No tests of omfit_github, because no guaranteed token')



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.sortedDict import SortedDict
from omfit_classes.utils_plot import View2d, View3d

import numpy as np

__all__ = ['OMFITuFile']


class OMFITuFile(SortedDict, OMFITascii):
    r"""
    Class used to interface with TRANSP U-files

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, **kw):
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        """
        Method used to load the content of the file specified in the .filename attribute

        :return: None
        """

        if self.filename is None or not os.stat(self.filename).st_size:
            self.reset()
            return

        with open(self.filename, 'r') as f:
            fl = f.readlines()

        # if file is empty, make bare bones structure
        if not len(fl):
            self.reset()
            return

        self['SHOT'] = int(re.sub(r'\s*([0-9]+)\s*\w+\s+[0-9]\s+[0-9]\s+[0-9]*.*', r'\1', fl[0]))
        self['DEVICE'] = re.sub(r'\s*[0-9]+\s*(\w+)\s+[0-9]\s+[0-9]\s+[0-9]*.*', r'\1', fl[0]).strip()
        self['NDIM'] = int(re.sub(r'\s*[0-9]+\s*\w+\s+([0-9])\s+[0-9]\s+[0-9]*.*', r'\1', fl[0]))
        self['COMPRESSION'] = int(re.sub(r'\s[0-9]+\s*\w+\s+[0-9]\s+([0-9])\s+[0-9]*.*', r'\1', fl[0]))
        try:
            self['LSHOT'] = max(6, int(re.sub(r'\s*[0-9]*\w+\s+[0-9]\s+[0-9]\s+([0-9]*).*', r'\1', fl[0])))
        except Exception:
            self['LSHOT'] = ''

        self['DATE'] = fl[1].split(';')[0].strip()
        nscal = int(fl[2].split(';')[0].strip())
        self['SCALARS'] = SortedDict()
        n = 2
        end_scalars = 3
        for k in range(nscal):
            value = float(fl[3 + 2 * k].split(';')[0].strip())
            name, desc = fl[3 + 2 * k + 1].split(';')[0].split(':')
            self['SCALARS'][k] = SortedDict()
            self['SCALARS'][k]['data'] = value
            self['SCALARS'][k]['name'] = name.strip()
            self['SCALARS'][k]['desc'] = desc.strip()
            n = 3 + 2 * k + 1
            end_scalars = 3 + 2 * k + 2

        # check dimensionality
        ndim = -1
        for i, line in enumerate(fl[end_scalars:]):
            try:
                int(line.split(';')[0].strip())
                break
            except Exception:
                ndim += 1

        # generic variable line reader
        def setvar(k, line):
            tmp = [line[:21].strip(), line[21:31].strip()]
            self[k] = {'name': tmp[0], 'units': tmp[1]}

        # label independent then dependent variables
        dims = []
        for i in range(ndim):
            setvar('X{:}'.format(i), fl[end_scalars + i])
            dims.append(int(fl[end_scalars + ndim + 2 + i].split(';')[0]))
        setvar('F', fl[end_scalars + ndim])

        # proc code
        self['PROCESSING_CODE'] = int(fl[end_scalars + ndim + 1].split(';')[0])

        # keep the whole header
        end_header = end_scalars + ndim * 2 + 2
        for i, line in enumerate(fl[:end_header]):
            self['__header%04d__' % i] = line.rstrip('\n')

        # find the trailer
        start_trailer = end_header
        for i, line in enumerate(fl[end_header:]):
            if 'END-OF-DATA' in line:
                start_trailer = end_header + i
                break

        # read data as array (TRANSP scruncher has non-critical bug where format leaves no space between negatives)
        flat = ' '.join(fl[end_header:start_trailer])
        flat = flat.replace('E-', 'E~').replace('e-', 'e~').replace('-', ' -').replace('E~', 'E-').replace('e~', 'e-')
        flat = np.array(flat.split(), dtype=float)
        for i in range(ndim):
            start = int(np.sum(dims[:i]))
            self['X{:}'.format(i)]['data'] = flat[start : start + dims[i]]
        # data read in fortran conventional order?
        self['F']['data'] = flat[int(np.sum(dims)) :].reshape(dims[::-1]).T

        self['COMMENTS'] = []
        for i, line in enumerate(fl[start_trailer + 1 :]):
            if 'Beams used:' in line:
                self['BEAMS'] = [_f for _f in [k.strip(', \n') for k in (line + fl[start_trailer + 2 + i]).split(' ')[3:]] if _f]
            if len(line.strip('\n')):
                self['COMMENTS'].append(line.rstrip('\n'))

    @dynaSave
    def save(self):
        """
        Save Ufile to the file specified in the .filename attribute

        :return: None
        """
        # write empty line if an empty ufile
        if not (self['SHOT'] and self['DEVICE'] and self['NDIM']):
            with open(self.filename, 'w') as f:
                f.write('')  # adds leading space to all lines
            return

        tmp = []
        if self['LSHOT']:
            tmp.append(
                '{:>6}{:<4} {:} {:} {:}              ;-SHOT #- F(X) DATA WRITEUF OMFIT'.format(
                    self['SHOT'], self['DEVICE'], self['NDIM'], self['COMPRESSION'], max(6, self['LSHOT'])
                )
            )
            fmt = '12.5e'
        else:
            tmp.append(
                '{:>6}{:<4} {:} {:}               ;-SHOT #- F(X) DATA WRITEUF OMFIT'.format(
                    self['SHOT'], self['DEVICE'], self['NDIM'], self['COMPRESSION']
                )
            )
            fmt = '11.4E'
        tmp.append('{:<11}                   ;-SHOT DATE-  UFILES ASCII FILE SYSTEM'.format(self['DATE']))
        tmp.append('{:3}                           ;-NUMBER OF ASSOCIATED SCALAR QUANTITIES-'.format(len(self['SCALARS'])))
        for k in list(self['SCALARS'].keys()):
            tmp.append('{:11.4E}                   ;-SCALAR, LABEL FOLLOWS:'.format(self['SCALARS'][k]['data']))
            tmp.append('{:10}{:20}'.format(self['SCALARS'][k]['name'] + ':', self['SCALARS'][k]['desc']))
        for i in range(len(self['F']['data'].shape)):
            k = 'X{:}'.format(i)
            tmp.append('{:<20}{:<10};-INDEPENDENT VARIABLE LABEL: {:}-'.format(self[k]['name'], self[k]['units'], k))
        tmp.append('{:<20}{:<10};-DEPENDENT VARIABLE LABEL-'.format(self['F']['name'], self['F']['units']))
        tmp.append('{:<3}                           ;-PROC CODE- 0:RAW 1:AVG 2:SM 3:AVG+SM'.format(self['PROCESSING_CODE']))
        for i in range(len(self['F']['data'].shape)):
            k = 'X{:}'.format(i)
            tmp.append('{:>10}                    ;-# OF X{:} PTS-'.format(len(self[k]['data']), i))
        tmp[-1].replace('PTS-', 'PTS- F(X) DATA FOLLOW:')

        for i in range(len(self['F']['data'].shape)):
            data = self['X{:}'.format(i)]['data']
            dlen = len(data)
            for i in range(0, dlen, 6):
                j = min(6, dlen - i)
                tmp.append((('{:' + fmt + '} ') * j).format(*data[i : i + j]).rstrip(' '))
        data = self['F']['data'].T.flatten()
        dlen = len(data)
        for i in range(0, dlen, 6):
            j = min(6, dlen - i)
            tmp.append((('{:' + fmt + '} ') * j).format(*data[i : i + j]).rstrip(' '))

        tmp.append(';----END-OF-DATA-----------------COMMENTS:-----------;')
        for line in tolist(self['COMMENTS']):
            tmp.append(line.strip())

        with open(self.filename, 'w') as f:
            f.write(' ' + '\n '.join(tmp) + '\n')  # adds leading space to all lines
        return

    @dynaLoad
    def plot(self, axes=None, figure=None, cmap=None, **kw):
        r"""
        Plot Ufile content

        :param axes: Axes object or None

        :param figure: Figure object or None

        :param cmap: Color map name used for multi-dimensional plots.

        :param \**kw: Extra key word arguments passed to matplotlib plot function.

        :return: Figure
        """
        if axes is None and figure is None:
            # Check if an empty figure was already created by the calling function to avoid 2 figures, one being empty
            if len(pyplot.get_fignums()) > 0 and len(pyplot.gcf().get_axes()) == 0:
                fig = pyplot.gcf()
                axes = fig.subplots()
            else:
                fig, axes = pyplot.subplots()
        elif axes is None and figure is not None:
            fig = figure
            if fig.axes:
                axes = fig.axes[0]
            else:
                pyplot.figure(num=fig.number)
                axes = subplot.subplot(111)
        elif axes is not None:
            fig = axes.get_figure()

        im_opt = {}
        if cmap is not None:
            im_opt['cmap'] = cmap

        xlabels = []
        for i in range(len(self['F']['data'].shape)):
            xlabels.append('{:} [{:}]'.format(self['X{:}'.format(i)]['name'], self['X{:}'.format(i)]['units']))
            xlabels[-1] = xlabels[-1].replace('[]', '')
        name = '{:} [{:}]'.format(self['F']['name'], self['F']['units']).replace(' []', '')

        # 1D data
        if len(np.squeeze(self['F']['data']).shape) == 1:
            kw.setdefault('label', name)
            kw.setdefault('marker', '.' * (len(self['X0']['data']) < 100))
            axes.plot(self['X0']['data'], self['F']['data'], **kw)
            axes.set_xlabel(xlabels[0])
            if self['F']['units'] == self['X0']['units'] and self['X0']['units']:
                axes.set_aspect('equal')
            leg = axes.legend()
            if leg:
                leg.draggable(True)

        # 2D data
        elif len(np.squeeze(self['F']['data']).shape) == 2:
            view = View2d(
                self['F']['data'],
                coords=[self['X0']['data'], self['X1']['data']],
                dims=xlabels,
                name=name.split()[-1],
                axes=axes,
                im_opt=im_opt,
                **kw,
            )
            view.colorbar.set_label(name)
            fig, axes = view.fig, view.axm
            fig.view = view

        # 3D data
        elif len(np.squeeze(self['F']['data']).shape) == 3:
            view = View3d(
                self['F']['data'],
                coords=[self['X0']['data'], self['X1']['data'], self['X2']['data']],
                dims=xlabels,
                name=name.split()[-1],
                axes=axes,
                im_opt=im_opt,
                **kw,
            )
            view.colorbar.set_label(name)
            fig, axes = view.fig, view.axm
            fig.view = view
        else:
            'Plotting not available above 3D'

        if 'BEAMS' in self:  # make a special beam overview
            fig2, axes = pyplot.subplots()
            for i in range(len(self['BEAMS'])):
                axes.plot(self['X0']['data'], self['F']['data'][:, i] / 1e6, label=self['BEAMS'][i])
            axes.plot(self['X0']['data'], np.sum(self['F']['data'][:, : len(self['BEAMS'])], 1) / 1e6, 'k--', label='Total')
            axes.set_xlabel('Time $[s]$')
            axes.set_ylabel('Power $[MW]$')
            axes.set_title('Neutral beams power from ' + os.path.split(self.filename)[1])
            leg = axes.legend(loc=0, frameon=False).draggable(True)
            # Additional info about the voltage, and current fractions
            fig_beam_quants, ax_beam_quants = pyplot.subplots(nrows=4, sharex=True)
            first_plot_lines = []
            quants = [('power [MW]', 1e-6), ('voltage [kV]', 1e-3), ('full frac', 1), ('half frac', 1)]
            for ind, beam in enumerate(self['BEAMS']):
                for jj, (name, mult) in enumerate(quants):
                    cur_ind = jj * len(self['BEAMS']) + ind
                    ll = ax_beam_quants[jj].plot(self['X0']['data'], self['F']['data'][:, cur_ind] * mult, label=beam)
                    if ind == 0:
                        ax_beam_quants[jj].set_ylabel(name)
                    if jj == 0:
                        first_plot_lines.append(ll)
            ax_beam_quants[-1].set_xlabel('Time [s]')
            # ax_beam_quants[0].legend(loc='best',frameon=False)
            ax_beam_quants[0].set_title('Neutral beams power from ' + os.path.split(self.filename)[1])
            fig_beam_quants.legend(first_plot_lines, labels=self['BEAMS'], loc='upper right', frameon=False)
            fig_beam_quants.subplots_adjust(right=0.8)
            for line in self['COMMENTS']:
                if 'tilt' in line:
                    printi(line)

        return fig

    @dynaLoad
    def smooth(self, window_x=None, window_len=11, window='hanning', axis=-1):
        """
        This built in function makes use of the OMFIT utils.smooth function
        to smooth over a single dimension of the data.

        If the axis in question is irregular, the data is first linearly interpolated onto
        a regular grid with spacing equal to the minimum step size of the irregular grid.

        :param window_x: Smoothing window size in axis coordinate units.
        :param window_len: Smoothing window size in index units. Ignored if window_x present. Enforced odd integer.
        :param window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
                flat window will produce a moving average smoothing.
        :param axis: Dimension over which to smooth. Accepts integer (0), key ('X0'), or name ('TIME').

        :return: OMFITuFile with the requested dimension smoothed

        """
        # generalize the axis specification
        if isinstance(axis, str):
            try:
                axis = int(axis.lstrip('X'))
            except Exception:
                for i in range(self['NDIM']):
                    if axis == self['X{:}'.format(i)]['name']:
                        axis = i
        if not isinstance(axis, int):
            raise ValueError('Unable to recognize axis {:}'.format(axis))
        axis = axis % (self['NDIM'])
        akey = 'X{:}'.format(axis)

        # copy if needed
        newu = copy.deepcopy(self)

        # force regular grid
        x = self[akey]['data']
        dx = np.diff(x)
        if not np.all(dx == dx[0]):
            printd('Interpolating to a regular grid', topic='OMFITuFile')
            newlen = int(np.ceil((x.max() - x.min()) / dx.min()))
            newx = np.linspace(x.min(), x.max(), newlen)
            fun = interp1d(x, self['F']['data'], axis=axis, kind='linear', bounds_error=False)
            newu[akey]['data'] = newx
            newu['F']['data'] = fun(newx)
            newu['COMMENTS'].append('Linearly interpolated to regular {:} grid.'.format(self[akey]['name']))
        else:
            newx = x

        # smooth
        if window_x:
            window_len = int(np.round(window_x / (newx[1] - newx[0]), 0))
        window_x = (newx[1] - newx[0]) * window_len
        newu['F']['data'] = utils.map1d(smooth, newu['F']['data'], axis=axis, window_len=window_len, window=window)

        # mark file as smoothed
        newu['PROCESSING_CODE'] = 2
        newu['COMMENTS'].append(
            'Smoothed over a {:} window with length of {:} in the {:} axis.'.format(window, window_x, newu[akey]['name'])
        )

        return newu

    def crop(self, xmin, xmax, axis=-1, endpoint=True):
        """
        Crop ufile data to only include points within the specified range along a given axis.
        Modifies the ufile object in place.

        :param xmin: Lower bound.
        :param xmax: Upper bound.
        :param axis: Dimension to bound. Accepts integer (0), key ('X0'), or name ('TIME').
        :param endpoint: Keeps an extra point on either end, ensuring the bounds are contained within the range of ufile data.

        """
        # generalize the axis specification
        if isinstance(axis, str):
            try:
                axis = int(axis.lstrip('X'))
            except Exception:
                for i in range(self['NDIM']):
                    if axis == self['X{:}'.format(i)]['name']:
                        axis = i
        if not isinstance(axis, int):
            raise ValueError('Unable to recognize axis {:}'.format(axis))
        axis = axis % (self['NDIM'])
        akey = 'X{:}'.format(axis)
        x = np.array(self[akey]['data'])
        # slice axis
        if endpoint:
            xslice = np.array([False] + list((x >= xmin) & (x <= xmax)) + [False])
            xslice = xslice[2:] | xslice[1:-1] | xslice[0:-2]
        else:
            xslice = (x >= xmin) & (x <= xmax)
        self[akey]['data'] = self[akey]['data'][xslice]
        # slice data
        dslice = [slice(stop) for i, stop in enumerate(np.shape(self['F']['data']))]
        dslice[axis] = xslice
        self['F']['data'] = self['F']['data'][dslice]
        # annotate
        self['COMMENTS'].append('Cropped range ({:},{:}) along {:} axis'.format(xmin, xmax, akey))

    def from_OMFITprofiles(self, OMFITprofiles_fit):
        """
        populate u-file based on OMFITprofiles fit xarray DataArray

        :params OMFITprofiles_fit: input OMFITprofiles fit xarray DataArray
        """
        self['NDIM'] = len(OMFITprofiles_fit.values.shape)
        self['SHOT'] = OMFITprofiles_fit.attrs['shot']
        self['DEVICE'] = OMFITprofiles_fit.attrs['device']

        self['DATE'] = now()
        self['COMMENTS'] = tolist('generated from OMFITprofiles')

        self['X0']['data'] = OMFITprofiles_fit['time'].values / 1000.0
        self['X0']['name'] = 'time'
        self['X0']['units'] = 's'

        self['F']['data'] = nominal_values(OMFITprofiles_fit.values / 1000.0)
        self['F']['name'] = OMFITprofiles_fit.name
        try:
            self['F']['units'] = OMFITprofiles_fit.attrs['units']
        except Exception:
            self['F']['units'] = ''

        if self['NDIM'] == 2:
            self['X1'] = dict()
            self['X1']['data'] = nominal_values(OMFITprofiles_fit[OMFITprofiles_fit.dims[1]].values)
            self['X1']['name'] = OMFITprofiles_fit.dims[1]
            try:
                self['X1']['units'] = OMFITprofiles_fit[OMFITprofiles_fit.dims[1]].attrs['units']
            except Exception:
                self['X1']['units'] = ''

    def reset(self):
        '''Set up basic bare bones structure'''
        self['SHOT'] = 0
        self['DEVICE'] = ''
        self['NDIM'] = 0
        self['COMPRESSION'] = 0
        self['LSHOT'] = 0
        self['DATE'] = ''
        self['SCALARS'] = SortedDict()
        self['X0'] = {'name': '', 'units': '', 'data': np.array([])}
        self['F'] = {'name': '', 'units': '', 'data': np.array([])}
        self['PROCESSING_CODE'] = 0
        self['COMMENTS'] = []



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_mds import OMFITmds
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_nc import OMFITncData
from omfit_classes.utils_base import printi
from omfit_classes.exceptions_omfit import OMFITexception
from omfit_classes.utils_math import interp1e
from omfit_classes.utils_fusion import is_device

import omas
import numpy as np
from scipy.interpolate import interp1d


__all__ = ['NoFitException', 'OMFITosborneProfile', 'OMFITpFile']

os.environ.setdefault('VPN_ACTIVE', '')


class NoFitException(OMFITexception):
    pass


class OMFITosborneProfile(OMFITmds):
    """
    Class accesses Osborne fits stored in MDSplus, and provides convenient methods
    for accessing the data or fitting functions used in those fits, including
    handling the covariance of the fitting parameters

    :param server: The device (really MDSplus archive)

    :param treename: The tree where the Osborne-tool profiles are stored

    :param shot: The shot number

    :param time: The timeid of the profile

    :param runid: The runid of the profile

    Note that this class assumes that the profiles are stored as
    [`tree`]['PROFDB_PED']['P<time>_<runid>']
    """

    def __init__(self, server='DIII-D', treename='auto', shot=None, time=None, runid=None):
        if treename == 'auto':
            if is_device(server, 'DIII-D'):
                treename = 'PEDESTAL'
            elif is_device(server, ('NSTX', 'NSTX-U')):
                treename = 'PROFDB_PED'
        self.OMFITproperties = {}
        OMFITmds.__init__(self, server, treename, shot)
        self.OMFITproperties['treename'] = treename
        self.OMFITproperties['time'] = time
        self.OMFITproperties['runid'] = runid
        self.OMFITproperties['shot'] = shot

    @property
    def treename(self):
        return self.OMFITproperties['treename']

    @treename.setter
    def treename(self, value):
        self.OMFITproperties['treename'] = value

    @property
    def time(self):
        return self.OMFITproperties['time']

    @time.setter
    def time(self, value):
        self.OMFITproperties['time'] = value

    @time.deleter
    def time(self):
        self.OMFITproperties['time'] = None

    @property
    def runid(self):
        return self.OMFITproperties['runid']

    @runid.setter
    def runid(self, value):
        self.OMFITproperties['runid'] = value

    @runid.deleter
    def runid(self):
        self.OMFITproperties['runid'] = None

    @property
    def shot(self):
        return self.OMFITproperties['shot']

    @shot.setter
    def shot(self, value):
        self.OMFITproperties['shot'] = value

    @shot.deleter
    def shot(self):
        self.OMFITproperties['shot'] = None

    @dynaLoad
    def load(self):
        OMFITmds.load(self)
        profdb_ped = self
        if self.treename == 'PEDESTAL':
            profdb_ped = self['PROFDB_PED']
        if 'P%s_%s' % (self.time, self.runid.upper()) not in profdb_ped:
            raise RuntimeError('Time: %s, Runid: %s does not exist for shot %s' % (self.time, self.runid, self.shot))
        tmp = profdb_ped['P%s_%s' % (self.time, self.runid.upper())]
        self.clear()
        self.update(tmp)

    def __repr__(self):
        return self.__class__.__name__ + '(' + ','.join(map(repr, [self.server, self.treename, self.shot, self.time, self.runid])) + ')'

    def get_raw_data(self, x_var='rhob', quant='ne'):
        """
        Get the raw data

        :param x_var: The mapping of the data (rhob,rhov,psi,R)

        :param quant: The quantity for which to return the data of the fit

        :return: x,y tuple * x - mapped radius of data
                           * y - the data (an array of numbers with uncertainties)
        """
        from uncertainties.unumpy import uarray

        node = self['%sDAT%s' % (quant.upper(), x_var.upper())]
        return node.dim_of(0), uarray(node.data(), node.error())

    def plot_raw_data(self, x_var='rhob', quant='ne', **kw):
        r"""
        Plot the raw data

        :param x_var: The mapping of the data (rhob,rhov,psi,R)

        :param quant: The quantity for which to plot the data of the fit

        :param \**kw: Keyword dictionary passed to uerrorbar

        :return: The collection instance returned by uerrorbar
        """
        from omfit_classes.utils_plot import uerrorbar
        from matplotlib import rcParams

        if quant == 'omeg':
            x, y = self.get_raw_data(quant='vt1', x_var=x_var)
            xR, yR = self.get_raw_data(quant='vt1', x_var='R')
            y = y / xR
        else:
            x, y = self.get_raw_data(x_var=x_var, quant=quant)
        kw.setdefault('color', rcParams['axes.prop_cycle'].by_key()['color'][0])
        kw.setdefault('ms', 0.5)
        kw.setdefault('capsize', 0)
        kw.setdefault('marker', '.')
        return uerrorbar(x, y, **kw)

    def calc_spline_fit(self, x_var='rhob', quant='ne', knots=5):
        """
        Calculate a spline fit

        :param x_var: The mapping of the data (rhob,rhov,psi,R)

        :param quant: The quantity for which to calculate a spline fit

        :param knots: An integer (autoknotted) or a list of knot locations
        """
        x, y = self.get_raw_data(x_var=x_var, quant=quant)
        fit = 'spl'
        fit_node = '%s%s%s' % (quant.upper(), fit.upper(), x_var.upper())
        print(fit_node in self)
        if fit_node in self:
            node = self[fit_node]
            xfit = node.dim_of(0)
            if xfit is not None:
                print(max(xfit), x < max(xfit))
                ind = x < max(xfit)
                x = x[ind]
                y = y[ind]
            if 'SPLINE_T' in node and node['SPLINE_T'].data() is not None:
                print('Using {fit_node} knots'.format(fit_node=fit_node))
                knots = np.unique(node['SPLINE_T'].data())
                print(knots)
        return fitSpline(x, nominal_values(y), std_devs(y), knots=knots, fit_SOL=True)

    def get_fit(self, fit, x_var='rhob', quant='ne', corr=True, x_val=None):
        """
        Get the fit, including uncertainties taking into account covariance of parameters

        :param fit: Which type of fit to retrieve

        :param x_var: The mapping of the fit (rhob,rhov,psi,R)

        :param quant: The quantity that was fit

        :param corr: (bool) Use the covariance of the parameters

        :param x_val: If the fit can be evaluated, if ``x_val`` is not ``None``,
            evaluate the fit at these locations

        :return: x,y tuple * x - radius of fit as stored in MDSplus
                           * y - the fit (an array of numbers with uncertainties)
        """
        node = self['%s%s%s' % (quant.upper(), fit.upper(), x_var.upper())]
        if node.data() is None:
            raise NoFitException('No fit for %s' % fit)

        def get_mdsplus_fit():
            if node.data().shape == node.error().shape:
                return node.dim_of(0), uarray(node.data(), node.error())
            else:
                return node.dim_of(0), uarray(node.data(), 0)

        import uncertainties
        import numpy as np
        from uncertainties.unumpy import uarray

        try:
            import data_fit_functions
        except Exception:
            warnings.warn('You must have Osborne tools installed to get the fit with covarianced errors.  \n' 'Using the stored errors')
            return get_mdsplus_fit()
        else:
            orig_np = np
            try:
                data_fit_functions.numpy = uncertainties.unumpy
                data_fit_functions.numpy.where = np.where
                try:
                    fit_fun = eval('data_fit_functions.%s' % (node['FITNAME'].data()[0]))
                except Exception:
                    warnings.warn("Can't handle fitname %s" % (node['FITNAME'].data()[0]))
                    return get_mdsplus_fit()
                if 'FIT_COEFCOV' not in node or not corr:
                    printi('Using Osborne tool profile fit parameters without correlated uncertainties for fit:%s' % fit)
                    params = uarray(node['FIT_COEF'].data(), node['FIT_COEFERR'].data())
                else:
                    params = uncertainties.correlated_values(node['FIT_COEF'].data(), node['FIT_COEFCOV'].data())
                try:
                    x = node.dim_of(0)
                    if x_val is not None:
                        x = x_val
                    if node['FITNAME'].data()[0] == 'tanh_multi' and x_var.upper() == 'RHOB':
                        return x, fit_fun(params, x, 0)
                    else:
                        return x, fit_fun(params, x)
                except Exception as _excp:
                    printe(_excp)
                    printe('Using MDSplus stored fit, without errorbars')
                    return get_mdsplus_fit()
            finally:
                data_fit_functions.numpy = orig_np

    def get_fit_deriv(self, nder, fit, **kw):
        r"""
        Apply deriv(x,y) `nder` times

        :param nder: Number of derivatives

        :param fit: Which type of fit to apply derivatives to

        :param \**kw: Keyword arguments passed to get_fit

        :return: x, d^{nder}y / d x^{nder} tuple * x - radius of fit as stored in MDSplus
                                                 * d^n y / d x^n - The nder'th derivative of the fit
        """
        x, y = self.get_fit(fit, **kw)
        for i in range(nder):
            y = deriv(x, y)
        return x, y

    def get_fit_params(self, fit, x_var='rhob', quant='ne', doc=False, covariance=True):
        """
        Get the parameters and their uncertainties

        :param fit: Which type of fit to retrieve (tnh0 is the correct tanhfit for scalings)

        :param x_var: The mapping of the fit (rhob,rhov,psi,R)

        :param quant: The quantity that was fit

        :param doc: (bool) if True print the fit documentation to understand which parameter is which

        :param covariance: (bool) return the covariance matrix instead of just the errors

        :return: params, cov -- tuple of parameters and their covariance matrix.
                Errors are the sqrt of diagonals: np.sqrt(np.diag(cov))
        """

        node = self['%s%s%s' % (quant.upper(), fit.upper(), x_var.upper())]
        if node.data() is None:
            raise NoFitException('No fit for %s' % fit)

        if doc:
            print(f"Fit documentation for {quant.upper()}{fit.upper()}{x_var.upper()}")
            print(node['FITDOC'].data()[0].format())

        if covariance:
            if 'FIT_COEFCOV' not in node:  # if the covariance matrix doesnt exist errors need to be in matrix
                printi('Covariance matrix not available using listed errors for fit:%s' % fit)
                params = node['FIT_COEF'].data()
                cov = np.zeros((len(params), len(params)))
                np.fill_diagnoal(cov, node['FIT_COEFERR'].data() ** 2)
            else:
                params, cov = node['FIT_COEF'].data(), node['FIT_COEFCOV'].data()
        else:  # if covariance is False just return the errors
            params = node['FIT_COEF'].data()
            cov = node['FIT_COEFERR'].data()

        return params, cov

    def plot_fit(self, fit, x_var='rhob', quant='ne', corr=True, nder=0, **kw):
        """
        Plot the fit for the given quant, with the given x-coordinate.

        :param fit: Which type of fit to plot

        :param x_var:  The mapping of the fit (rhob,rhov,psi,R)

        :param quant: The quantity that was fit

        :param corr: Plot with correlated uncertainty parameters

        :param nder: The number of derivatives to compute before plotting

        :return: The result of uband
        """
        x, y = self.get_fit_deriv(nder, fit, x_var=x_var, quant=quant, corr=corr)
        from omfit_classes.utils_plot import uband

        return uband(x, y, **kw)

    def plot_all_fits(self, x_var='rhob', quant='ne', corr=True, nder=0):
        """
        Plot all fits for the given quantity `quant` and mapping `x_var`

        :param x_var: The mapping of the fit (rhob,rhov,psi,R)

        :param quant: The quantity that was fit

        :param corr: Plot with correlated uncertainty parameters

        :param nder: Plot the nder'th derivative of the fit (if nder=0, plot data also)

        :return: None
        """
        if nder == 0:
            self.plot_raw_data(x_var=x_var, quant=quant)
        ki = 1
        for k in list(self.keys()):
            if not (k.startswith(quant.upper()) and k.endswith(x_var.upper())):
                continue
            fit_type = k[len(quant) : -len(x_var)]
            if fit_type == 'DAT':
                continue
            try:
                from matplotlib import rcParams

                self.plot_fit(
                    fit_type,
                    x_var=x_var,
                    quant=quant,
                    color=rcParams['axes.prop_cycle'].by_key()['color'][ki % len(rcParams['axes.prop_cycle'].by_key()['color'])],
                    corr=corr,
                    label=fit_type,
                    nder=nder,
                )
                ki += 1
            except (NoFitException, TypeError) as _excp:
                continue
        from pylab import draw

        draw()

    def __save_kw__(self):
        """
        :return: generate the kw dictionary used to save the attributes to be passed when reloading from OMFITsave.txt
        """
        return self.OMFITproperties


class OMFITpFile(SortedDict, OMFITascii):
    r"""
    OMFIT class used to interface with Osborne pfiles

    :param filename: filename passed to OMFITobject class

    :param \**kw: keyword dictionary passed to OMFITobject class
    """

    def __init__(self, filename, **kw):
        OMFITobject.__init__(self, filename, **kw)
        SortedDict.__init__(self)

        self.legacy = False
        self.is_remapped = False  # define whether this is the original pFile or its psin grid has been remapped

        self.descriptions = SortedDict()
        self.descriptions['ne'] = 'Electron density'
        self.descriptions['te'] = 'Electron temperature'
        self.descriptions['ni'] = 'Ion density'
        self.descriptions['ti'] = 'Ion temperature'
        self.descriptions['nb'] = 'Fast ion density'
        self.descriptions['pb'] = 'Fast ion pressure'
        self.descriptions['ptot'] = 'Total pressure'
        self.descriptions['omeg'] = 'Toroidal rotation: VTOR/R'
        self.descriptions['omegp'] = 'Poloidal rotation: Bt * VPOL / (RBp)'
        self.descriptions['omgvb'] = 'VxB rotation term in the ExB rotation frequency: OMEG + OMEGP'
        self.descriptions['omgpp'] = 'Diamagnetic term in the ExB rotation frequency: (P_Carbon)/dpsi / (6*n_Carbon)'
        self.descriptions['omgeb'] = 'ExB rotation frequency: OMGPP + OMGVB = Er/(RBp)'
        self.descriptions['er'] = 'Radial electric field from force balance: OMGEB * RBp'
        self.descriptions['ommvb'] = 'Main ion VXB term of Er/RBp, considered a flux function'
        self.descriptions['ommpp'] = 'Main ion pressure term of Er/RBp, considered a flux function'
        self.descriptions['omevb'] = 'Electron VXB term of Er/RBp, considered a flux function'
        self.descriptions['omepp'] = 'Electron pressure term of Er/RBp, considered a flux function'
        self.descriptions['kpol'] = 'KPOL=VPOL/Bp : V_vector = KPOL*B_vector + OMGEB * PHI_Vector'
        self.descriptions['N Z A'] = 'N Z A of ION SPECIES'
        self.descriptions['omghb'] = 'Hahm-Burrell form for the ExB velocity shearing rate: OMGHB = (RBp)**2/Bt * d (Er/RBp)/dpsi'
        self.descriptions['nz1'] = 'Density of the 1st impurity species'
        self.descriptions['vtor1'] = 'Toroidal velocity of the 1st impurity species'
        self.descriptions['vpol1'] = 'Poloidal velocity of the 1st impurity species'
        self.descriptions['nz2'] = 'Density of the 2nd impurity species'
        self.descriptions['vtor2'] = 'Toroidal velocity of the 2nd impurity species'
        self.descriptions['vpol2'] = 'Poloidal velocity of the 2nd impurity species'
        # There may be more impurity species, but let's stop here for now.

        self.units = SortedDict()
        self.units['ne'] = '10^20/m^3'
        self.units['te'] = 'KeV'
        self.units['ni'] = '10^20/m^3'
        self.units['ti'] = 'KeV'
        self.units['nb'] = '10^20/m^3'
        self.units['pb'] = 'KPa'
        self.units['ptot'] = 'KPa'
        self.units['omeg'] = 'kRad/s'
        self.units['omegp'] = 'kRad/s'
        self.units['omgvb'] = 'kRad/s'
        self.units['omgpp'] = 'kRad/s'
        self.units['omgeb'] = 'kRad/s'
        self.units['ommvb'] = ''
        self.units['ommpp'] = ''
        self.units['omevb'] = ''
        self.units['omepp'] = ''
        self.units['er'] = 'kV/m'
        self.units['kpol'] = 'km/s/T'
        self.units['N Z A'] = ''
        self.units['omghb'] = ''
        self.units['nz1'] = '10^20/m^3'
        self.units['vtor1'] = 'km/s'
        self.units['vpol1'] = 'km/s'
        self.units['nz2'] = '10^20/m^3'
        self.units['vtor2'] = 'km/s'
        self.units['vpol2'] = 'km/s'

        if os.stat(self.filename).st_size:
            self.dynaLoad = True
        else:
            for key in list(self.descriptions.keys()):
                if key in ('N Z A',):
                    continue
                self[key] = OMFITncData()
                self[key]['data'] = np.array([0])
                self[key]['description'] = self.descriptions[key]
                self[key]['psinorm'] = np.array([0])
                self[key]['units'] = self.units[key]
                self[key]['derivative'] = np.array([0])

    @dynaLoad
    def load(self):
        """
        Method used to load the content of the file specified in the .filename attribute

        :return: None
        """

        self.clear()

        # read the file
        f = os.path.getsize(self.filename)
        if f == 0:
            return
        fn = self.filename
        with open(fn, 'r') as f:
            fl = f.read().strip().splitlines(True)

        ind = 0
        while True:
            try:
                line = fl[ind]
                cur = line.split()
                curData = fl[ind + 1].split()
            except IndexError:
                break

            num = int(re.sub(r'([0-9]*)\s.*', r'\1', line))
            xkey = re.sub(r'([0-9]*)\s(.*)\s(.*)\((.*)\)\s(.*)\n', r'\2', line)
            key = re.sub(r'([0-9]*)\s(.*)\s(.*)\((.*)\)\s(.*)\n', r'\3', line)
            units = re.sub(r'([0-9]*)\s(.*)\s(.*)\((.*)\)\s(.*)\n', r'\4', line)
            der = re.sub(r'([0-9]*)\s(.*)\s(.*)\((.*)\)\s(.*)\n', r'\5', line)

            quants = [xkey, key + '(' + units + ')', der]
            q = []
            for i in range(ind + 1, ind + 1 + num):
                q.append(list(map(float, fl[i].split())))
            q = list(zip(*q))
            if key in self:
                if np.sum(np.array(self[key]['data']) != np.array(q[1])):
                    raise ValueError('%s already defined, but trying to change its value' % quants[1])
            elif 'N Z A of ION SPECIES' in line:
                self['N Z A'] = OMFITncData()
                self['N Z A']['description'] = 'N Z A of ION SPECIES'
                self['N Z A']['N'] = np.array(q[0])
                self['N Z A']['Z'] = np.array(q[1])
                self['N Z A']['A'] = np.array(q[2])
            else:
                self[key] = OMFITncData()
                if key in self.descriptions:
                    self[key]['description'] = self.descriptions[key]
                else:
                    self[key]['description'] = key
                self[key]['units'] = units
                for qi, quant in enumerate(quants):
                    if qi == 1:
                        self[key]['data'] = np.array(q[qi])
                    elif quant == 'd%s/dpsiN' % key or quant == 'd%s/dpsi' % key:
                        self[key]['derivative'] = np.array(q[qi])
                    else:
                        self[key][quant] = np.array(q[qi])

            ind = ind + 1 + num

    @dynaSave
    def save(self):
        """
        Method used to save the content of the object to the file specified in the .filename attribute

        :return: None
        """
        lines = []
        for key in self.keys(filter=hide_ptrn):
            if key == 'N Z A':
                if self.legacy:
                    break
                lines.append('%i N Z A of ION SPECIES\n' % (len(self[key]['A']),))
                for i in range(len(self[key]['A'])):
                    lines.append(" %f   %f   %f\n" % (self[key]['N'][i], self[key]['Z'][i], self[key]['A'][i]))
            else:
                if len(self[key]['data']) == 1:
                    continue
                lines.append("%i psinorm %s(%s) d%s/dpsiN\n" % (len(self[key]['data']), key, self[key]['units'], key))
                for i in range(len(self[key]['data'])):
                    lines.append(" %f   %f   %f\n" % (self[key]['psinorm'][i], self[key]['data'][i], self[key]['derivative'][i]))
        with open(self.filename, 'w') as f:
            f.writelines(lines)

    def plot(self):
        """
        Method used to plot all profiles, one in a different subplots

        :return: None
        """
        nplot = int(len(self))
        cplot = int(np.floor(np.sqrt(nplot) / 2.0) * 2)
        rplot = int(np.ceil(nplot * 1.0 / cplot))
        pyplot.subplots_adjust(wspace=0.35, hspace=0.0)
        pyplot.ioff()
        try:
            for k, item in enumerate(self):
                r = int(np.floor(k * 1.0 / cplot))
                c = int(k - r * cplot)

                if k == 0:
                    ax1 = pyplot.subplot(rplot, cplot, r * (cplot) + c + 1)
                    ax = ax1
                else:
                    ax = pyplot.subplot(rplot, cplot, r * (cplot) + c + 1, sharex=ax1)
                ax.ticklabel_format(style='sci', scilimits=(-3, 3))
                if 'psinorm' not in self[item]:
                    printi('Can\'t plot %s because no psinorm attribute' % item)
                    continue
                x = self[item]['psinorm']

                pyplot.plot(x, self[item]['data'], '.-')
                pyplot.text(0.5, 0.9, item, horizontalalignment='center', verticalalignment='top', size='medium', transform=ax.transAxes)

                if k < len(self) - cplot:
                    pyplot.setp(ax.get_xticklabels(), visible=False)
                else:
                    pyplot.xlabel('$\\psi$')

                pyplot.xlim(min(x), max(x))

        finally:
            pyplot.draw()
            pyplot.ion()

    def remap(self, points='ne', **kw):
        """
        Remap the disparate psinorm grids for each vaiable onto the same psinorm grid

        :param npoints: number of points to remap the originally 256 grid
            - If points is int: make an evenly spaced array.
            - If points is array: use this as the grid
            - If points is a string: use the 'psinorm' from that item in the pfile (by default `ne['psinorm']`)

        :param \**kw: addidional keywords are passed to scipy.interpolate.interp1d

        return: The entire object remapped to the same psinorm grid
        """

        mapped = copy.deepcopy(self)
        keys = mapped.keys()

        if isinstance(points, str):
            remap_psin = self[points]['psinorm']

        elif type(points) is int:  # Evenly spaced grid rebase
            remap_psin = np.linspace(0, 1, points)

        elif type(points) in [list, tuple, np.ndarray]:  # user provided grid rebase
            assert np.all(np.abs(points) <= 1), 'All values must be on the interval [0, 1]'
            assert np.all(np.diff(points) > 0), 'All values must be monotonically increasing'
            remap_psin = points

        mapped['psinorm'] = remap_psin  # save the psin grid
        # iterate through all keys
        for key in keys:
            if key in ('N Z A',):  # this is not mapped to psi so skip it
                continue
            f_key = interp1d(self[key]['psinorm'], self[key]['data'])
            f_key_derivative = interp1d(self[key]['psinorm'], self[key]['derivative'], **kw)
            mapped[key]['psinorm'] = remap_psin
            mapped[key]['data'] = f_key(remap_psin)
            mapped[key]['derivative'] = f_key_derivative(remap_psin)

        return mapped

    def to_omas(self, ods=None, time_index=0, gEQDSK=None):
        """
        translate OMFITpFile class to OMAS data structure

        :param ods: input ods to which data is added

        :param time_index: time index to which data is added

        :param gEQDSK: corresponding gEQDSK (if ods does not have equilibrium already

        :return: ods
        """
        # Lyons:  I've used my own rotation variable definitions here, until IMAS properly defines them
        #         Note that negative signs and sign_Ip are needed to make
        #           omega_perp, omega_dia, and omega_ExB have the same sign convention

        from omfit_classes.omfit_eqdsk import OMFITgeqdsk
        from omas import ODS, omas_environment
        from omfit_classes.utils_math import atomic_element

        if ods is None:
            ods = ODS()

        if gEQDSK is not None:
            gEQDSK.to_omas(ods=ods, time_index=time_index)
        else:
            filename = 'g' + os.path.split(self.filename)[1][1:]
            gEQDSK = OMFITgeqdsk(filename).from_omas(ods=ods, time_index=time_index)

        cocosio = 1  # pFile's have CCW phi and CW theta
        assert gEQDSK.cocos == cocosio, "gEQDSK must be in COCOS %d" % cocosio

        # get required quantities
        sign_Ip = np.sign(gEQDSK['CURRENT'])

        # get gEQDSK quantities
        psin_eq = np.linspace(0.0, 1.0, len(gEQDSK['PRES']))
        rhotn_eq = gEQDSK['RHOVN']
        R_eq = gEQDSK['fluxSurfaces']['midplane']['R']
        Bt_eq = gEQDSK['fluxSurfaces']['midplane']['Bt']
        Bp_eq = gEQDSK['fluxSurfaces']['midplane']['Bp']

        quants = [
            ('rotation_frequency_tor_sonic', 'omgeb', sign_Ip * 1e3),  # sign_Ip accounts for abs(Bp) in pFile definition
            ('pressure_perpendicular', 'ptot', 1e3 / 3.0),
            ('pressure_parallel', 'ptot', 1e3 / 3.0),
            ('e_field.radial', 'er', 1e3),
            ('electrons.density_thermal', 'ne', 1e20),
            ('electrons.temperature', 'te', 1e3),
            ('electrons.rotation.perpendicular', 'omevb', 1e3),
            ('electrons.rotation.diamagnetic', 'omepp', sign_Ip * 1e3),
        ]  # sign_Ip accounts for abs(dpsi) in pFile definition

        def get_ion(N, Z, A):
            label = list(atomic_element(Z=N, A=int(A)).values())[0]['symbol']
            k = -1
            ion = None
            prof1d = ods['core_profiles']['profiles_1d'][time_index]
            if 'ion' in prof1d:
                for k in prof1d['ion']:
                    if (
                        prof1d['ion'][k]['label'] == label
                        and int(prof1d['ion'][k]['element'][0]['a']) == int(A)
                        and prof1d['ion'][k]['element'][0]['z_n'] == Z
                    ):
                        # found the ion
                        ion = prof1d['ion'][k]
                        err = "OMAS only works for single ion states"
                        assert ion['multiple_states_flag'] == 0, err
            if ion is None:
                k += 1
                # need to add new ion
                ion = prof1d['ion'][k]
                ion['multiple_states_flag'] = 0
                ion['label'] = label
                ion['element'][0]['a'] = A
                ion['element'][0]['z_n'] = Z

            return k, ion

        if 'N Z A' in self:
            # Assume multiple ion species are defined
            mk, main_ion = get_ion(N=self['N Z A']['N'][-2], Z=self['N Z A']['Z'][-2], A=self['N Z A']['A'][-2])
            bk, beam_ion = get_ion(N=self['N Z A']['N'][-1], Z=self['N Z A']['Z'][-1], A=self['N Z A']['A'][-1])
        else:
            printw("'N Z A' missing from OMFITpFile; assuming deuterium for main and beam ions")
            mk, main_ion = get_ion(N=1, Z=1, A=2.0)
            bk, beam_ion = get_ion(N=1, Z=1, A=2.0)

        quants += [
            (f'ion.{mk}.density_thermal', 'ni', 1e20),
            (f'ion.{mk}.temperature', 'ti', 1e3),
            (f'ion.{mk}.rotation.perpendicular', 'ommvb', 1e3),
            (f'ion.{mk}.rotation.diamagnetic', 'ommpp', sign_Ip * 1e3),  # sign_Ip accounts for abs(dpsi) in pFile definition
            (f'ion.{bk}.density_fast', 'nb', 1e20),
            (f'ion.{bk}.pressure_fast_perpendicular', 'pb', 1e3 / 3.0),
            (f'ion.{bk}.pressure_fast_parallel', 'pb', 1e3 / 3.0),
            (f'ion.{mk}.velocity.toroidal', 'vtor1', 1e3),
            (f'ion.{mk}.velocity.poloidal', 'vpol1', 1e3),
        ]

        if 'N Z A' in self:
            for i in range(len(self['N Z A']['N']) - 2):
                ik, imp_ion = get_ion(N=self['N Z A']['N'][i], Z=self['N Z A']['Z'][i], A=self['N Z A']['A'][i])
                quants += [
                    (f'ion.{ik}.density_thermal', 'nz%d' % (i + 1), 1e20),
                    (f'ion.{ik}.temperature', 'ti', 1e3),
                ]  # Lyons: I think it's assumed all ions have same temperature

                if imp_ion['label'] == 'C':
                    # working with Carbon
                    quants += [
                        (f'ion.{ik}.rotation.perpendicular', 'omgvb', 1e3),
                        (f'ion.{ik}.rotation.diamagnetic', 'omgpp', sign_Ip * 1e3),  # sign_Ip accounts for abs(dpsi) in pFile definition
                        (f'ion.{ik}.rotation.parallel_stream_function', 'kpol', 1e3),
                    ]

        else:
            ik, imp_ion = get_ion(N=6, Z=6, A=12.0)
            if 'nz1' in self:
                printw("                                 carbon impurity")
                quants += [(f'ion.{ik}.density_thermal', 'nz1', 1e20)]
            else:
                printw("                                 carbon impurity with trace density")
                quants += [(f'ion.{ik}.density_thermal', 'ni', 1e-6)]
            quants += [
                (f'ion.{ik}.temperature', 'ti', 1e3),  # Lyons: I think it's assumed all ions have same temperature
                (f'ion.{ik}.rotation.perpendicular', 'omgvb', 1e3),
                (f'ion.{ik}.rotation.diamagnetic', 'omgpp', sign_Ip * 1e3),  # sign_Ip accounts for abs(dpsi) in pFile definition
                (f'ion.{ik}.rotation.parallel_stream_function', 'kpol', 1e3),
            ]

        for q_ods, q_p, fac in quants:
            if q_p in self:
                # pFiles don't enforce that every variable has the same psinorm grid
                rho_tor_norm = np.interp(self[q_p]['psinorm'], psin_eq, rhotn_eq)
                coordsio = {f'core_profiles.profiles_1d.{time_index}.grid.rho_tor_norm': rho_tor_norm}
                with omas_environment(ods, cocosio=cocosio, coordsio=coordsio):
                    prof1d = ods['core_profiles']['profiles_1d'][time_index]
                    prof1d[q_ods] = fac * self[q_p]['data']
            else:
                printw("pFile does not contain " + q_p + ". Skipping")

        if 'N Z A' in self:
            for i in range(len(self['N Z A']['N']) - 2):
                ik, imp_ion = get_ion(N=self['N Z A']['N'][i], Z=self['N Z A']['Z'][i], A=self['N Z A']['A'][i])
                if imp_ion['label'] != 'C':
                    try:
                        # save rotations for non-carbon
                        psin = self[f'vpol{(i + 1)}']['psinorm']
                        rho_tor_norm = np.interp(psin, psin_eq, rhotn_eq)
                        coordsio = {f'core_profiles.profiles_1d.{time_index}.grid.rho_tor_norm': rho_tor_norm}

                        R = np.interp(psin, psin_eq, R_eq)
                        Bp = np.interp(psin, psin_eq, Bp_eq)
                        Bt = np.interp(psin, psin_eq, Bt_eq)

                        with omas_environment(ods, cocosio=cocosio, coordsio=coordsio):
                            prof1d = ods['core_profiles']['profiles_1d'][time_index]

                            # parallel stream function from vpol and Bp
                            vpol = sign_Ip * self[f'vpol{(i + 1)}']['psinorm']['data']  # vpol in pFile has a sign_Ip
                            prof1d[f'ion.{ik}.velocity.poloidal'] = 1e3 * vpol
                            prof1d[f'ion.{ik}.rotation.parallel_stream_function'] = 1e3 * vpol / Bp

                            # perpendicular rotation from vtor and vpol
                            vtor = np.interp(psin, self[f'vtor{(i + 1)}']['psinorm'], self[f'vtor{(i + 1)}']['data'])
                            prof1d[f'ion.{ik}.velocity.toroidal'] = 1e3 * vtor
                            omeg = vtor / R
                            omegp = -vpol * Bt / (R * Bp)
                            prof1d[f'ion.{ik}.rotation.perpendicular'] = 1e3 * (omeg + omegp)

                            # diamagentic rotation from vtor, vpol, and omgeb
                            omgeb = sign_Ip * np.interp(
                                psin, self['omgeb']['psinorm'], self['omgeb']['data']
                            )  # sign_Ip accounts for abs(Bp) in pFile definition
                            prof1d[f'ion.{ik}.rotation.diamagnetic'] = 1e3 * (omgeb - omeg - omegp)

                    except KeyError:
                        printw(f'Some velocity info missing for pFile ion {(i + 1)}, so possible missing info for ODS ion {ik}')
                        continue

        return ods

    def from_omas(self, ods=None, time_index=0):
        """
        translate OMAS data structure to OMFITpFile

        :param ods: input ods to take data from

        :param time_index: time index to which data is added

        :return: ods
        """
        # Lyons:  I've used my own rotation variable definitions here, until IMAS properly defines them
        #         Note that negative signs and sign_Ip are needed to make
        #           omega_perp, omega_dia, and omega_ExB have the same sign convention

        from omfit_classes.omfit_eqdsk import OMFITgeqdsk
        from omas import ODS, omas_environment
        from omfit_classes.utils_math import atomic_element

        cocosio = 1  # pFile's have CCW phi and CW theta

        # get required quantities

        with omas_environment(ods, cocosio=cocosio):
            sign_Ip = np.sign(ods['equilibrium']['time_slice'][time_index]['global_quantities']['ip'])

        prof1d = ods['core_profiles']['profiles_1d'][time_index]

        # determine main ion by highest mean thermal density
        den_mk = 0.0
        ions = list(prof1d['ion'].keys())
        ck = None
        for k in ions:
            den_k = np.mean(prof1d['ion'][k]['density_thermal'])
            if den_k > den_mk:
                # define main ion
                den_mk = den_k
                mk = k
            if prof1d['ion'][k]['label'] == 'C':
                # define carbon ion (if found)
                ck = k

        # move main ion to end
        ions.remove(mk)
        ions += [mk]

        # move carbon ion to the beginning
        if ck is not None:
            ions.remove(ck)
            ions = [ck] + ions

        Nions = len(ions) + 1
        self['N Z A'] = OMFITncData()
        self['N Z A']['description'] = 'N Z A of ION SPECIES'
        self['N Z A']['N'] = np.zeros(Nions)
        self['N Z A']['Z'] = np.zeros(Nions)
        self['N Z A']['A'] = np.zeros(Nions)

        quants = [
            ('ne', 'electrons.density_thermal', 1e-20),
            ('te', 'electrons.temperature', 1e-3),
            ('ni', f'ion.{mk}.density_thermal', 1e-20),
            ('ti', f'ion.{mk}.temperature', 1e-3),
            ('nb', f'ion.{mk}.density_fast', 1e-20),
            ('pb', [f'ion.{mk}.pressure_fast_perpendicular', f'ion.{mk}.pressure_fast_parallel'], [2e-3, 1e-3]),
            ('ptot', ['pressure_perpendicular', 'pressure_parallel'], [2e-3, 1e-3]),
            ('omgeb', 'omega0', sign_Ip * 1e-3),  # sign_Ip accounts for abs(Bp) in pFile definition
            ('omevb', 'electrons.rotation.perpendicular', 1e-3),
            ('omepp', 'electrons.rotation.diamagnetic', sign_Ip * 1e-3),  # sign_Ip accounts for abs(dpsi) in pFile definition
            ('ommvb', f'ion.{mk}.rotation.perpendicular', 1e-3),
            ('ommpp', f'ion.{mk}.rotation.diamagnetic', sign_Ip * 1e-3),
        ]  # sign_Ip accounts for abs(dpsi) in pFile definition

        for i, k in enumerate(ions):
            # currently we don't store charge state information, so Z = N
            N = prof1d['ion'][k]['element'][0]['z_n']
            A = prof1d['ion'][k]['element'][0]['a']
            if k == mk:
                # main and beam ions
                self['N Z A']['N'][-1] = self['N Z A']['N'][-2] = N
                self['N Z A']['Z'][-1] = self['N Z A']['Z'][-2] = N
                self['N Z A']['A'][-1] = self['N Z A']['A'][-2] = A
            else:
                # impurity ions
                self['N Z A']['N'][i] = N
                self['N Z A']['Z'][i] = N
                self['N Z A']['A'][i] = A

                pk = i + 1  # iterate and use for pfile indexing
                if f'nz{pk}' not in self:
                    for key, desc, units in [
                        (f'nz{pk}', 'Density', '10^20/m^3'),
                        (f'vtor{pk}', 'Toroidal velocity', 'km/s'),
                        (f'vpol{pk}', 'Toroidal velocity', 'km/s'),
                    ]:
                        self[key] = OMFITncData()
                        self[key]['data'] = np.array([0])
                        self[key]['description'] = desc + f' of the #{pk} impurity species'
                        self[key]['psinorm'] = np.array([0])
                        self[key]['units'] = units
                        self[key]['derivative'] = np.array([0])

                quants += [(f'nz{pk}', f'ion.{k}.density_thermal', 1e-20)]

        # load gEQDSK from ODS
        # Need to identify correct time.
        gEQDSK = OMFITgeqdsk('g0.0').from_omas(ods, time_index=time_index)
        gtime = ods[f'equilibrium.time_slice.{time_index}.time']
        if gtime != prof1d['time']:
            printw("Warning: The time corresponding to the equilibrium and the profile is mismatched!")
            printw(f"    Please check your input ods!")
            printw(
                f"    For time_index = {time_index}, the equilibrium has time = {gtime}, and the core_profile has time = {prof1d['time']}."
            )

        # load data from ODS
        with omas_environment(ods, cocosio=cocosio):
            psi1D = interp1e(
                ods[f'equilibrium.time_slice.{time_index}.profiles_1d.rho_tor_norm'],
                ods[f'equilibrium.time_slice.{time_index}.profiles_1d.psi'],
            )(ods[f'core_profiles.profiles_1d.{time_index}.grid.rho_tor_norm'])
            coordsio = {f'equilibrium.time_slice.{time_index}.profiles_1d.psi': psi1D}

        with omas_environment(ods, cocosio=cocosio, coordsio=coordsio):
            prof1d = ods['core_profiles']['profiles_1d'][time_index]
            # pFiles are on psinorm grid but core_profiles data on rho_tor_norm gird
            rhotn = ods['equilibrium']['time_slice'][time_index]['profiles_1d']['rho_tor_norm']
            psi = ods['equilibrium']['time_slice'][time_index]['profiles_1d']['psi']
            psib = np.interp(1.0, rhotn, psi)  # in case grid extends beyond rhotn = 1.
            psin = (psi - psi[0]) / (psib - psi[0])

            for q_p, q_ods, fac in quants:
                self[q_p]['psinorm'] = psin
                try:
                    if isinstance(q_ods, list):
                        self[q_p]['data'] = 0.0 * psin
                        for i, qo in enumerate(q_ods):
                            self[q_p]['data'] += fac[i] * nominal_values(prof1d[qo])
                    else:
                        self[q_p]['data'] = fac * nominal_values(prof1d[q_ods])
                except ValueError:
                    self[q_p]['data'] = 0.0 * psin

            # calc auxiliary quantities from gEQDSK['fluxSurfaces']
            psin_eq = np.linspace(0.0, 1.0, len(gEQDSK['PRES']))
            R = np.interp(psin, psin_eq, gEQDSK['fluxSurfaces']['midplane']['R'])
            Bt = np.interp(psin, psin_eq, gEQDSK['fluxSurfaces']['midplane']['Bt'])
            Bp = np.interp(psin, psin_eq, gEQDSK['fluxSurfaces']['midplane']['Bp'])

            self['er']['psinorm'] = psin
            self['er']['data'] = R * abs(Bp) * self['omgeb']['data']  # pFile Er same sign as omgeb

            zero_array = 0.0 * psin

            for ik, k in enumerate(ions):
                ik += 1
                if k != mk:
                    # these are all as they're defined in pFile
                    kpol = None
                    omegp = None
                    vpol = None
                    omgvb = None
                    omeg = None
                    vtor = None
                    omgpp = None
                    missing = []

                    if f'ion.{k}.rotation.parallel_stream_function' in prof1d:
                        kpol = 1e-3 * nominal_values(prof1d[f'ion.{k}.rotation.parallel_stream_function'])
                        omegp = -Bt * kpol / R
                        vpol = sign_Ip * kpol * Bp  # sign_Ip accounts for pFile definition

                    if f'ion.{k}.rotation_frequency_tor' in prof1d:
                        omeg = 1e-3 * nominal_values(prof1d[f'ion.{k}.rotation_frequency_tor'])

                    if f'ion.{k}.rotation.perpendicular' in prof1d:
                        omgvb = 1e-3 * nominal_values(prof1d[f'ion.{k}.rotation.perpendicular'])

                    if f'ion.{k}.rotation.diamagnetic' in prof1d:
                        # sign_Ip accounts for abs(dpsi) in pFile definition
                        omgpp = -1e-3 * sign_Ip * nominal_values(prof1d[f'ion.{k}.rotation.diamagnetic'])

                    omgeb = self['omgeb']['data']
                    if (omgvb is None) and (omeg is not None) and (omegp is not None):
                        omgvb = omeg + omegp

                    if (not np.any(omgeb)) and (omgpp is not None) and (omgvb is not None):
                        omgeb = omgpp + sign_Ip * omgvb

                    if omeg is None and omgvb is not None and omegp is not None:
                        omeg = omgvb - omegp

                    if omeg is not None:
                        vtor = R * omeg
                        if omgpp is None:
                            # sign_Ip accounts pFile definition
                            omgpp = omgeb - sign_Ip * omgvb

                    if vpol is None and f'ion.{k}.velocity.poloidal' in prof1d:
                        vpol = 1e-3 * nominal_values(prof1d[f'ion.{k}.velocity.poloidal'])

                    if vtor is None and f'ion.{k}.velocity.toroidal' in prof1d:
                        vtor = 1e-3 * nominal_values(prof1d[f'ion.{k}.velocity.toroidal'])

                    # Calculate fallback omevb and ommvb
                    omevb = self['omevb']['data']  # These are in params above and will be present, but may be all zero
                    omepp = self['omepp']['data']
                    ommvb = self['ommvb']['data']
                    ommpp = self['ommpp']['data']

                    if (not np.any(omevb)) and (omepp is not None) and np.any(omgeb):
                        omevb = sign_Ip * (omgeb - omepp)
                    if (not np.any(ommvb)) and (ommpp is not None) and np.any(omgeb):
                        ommvb = sign_Ip * (omgeb - ommpp)

                    params = [(f'vpol{ik}', vpol), (f'vtor{ik}', vtor)]
                    if ik == 1:
                        params += [
                            ('kpol', kpol),
                            ('omegp', omegp),
                            ('omgvb', omgvb),
                            ('omeg', omeg),
                            ('omgpp', omgpp),
                            ('omevb', omevb),
                            ('ommvb', ommvb),
                            ('omgeb', omgeb),
                        ]

                    for name, val in params:
                        self[name]['psinorm'] = psin
                        if val is not None:
                            self[name]['data'] = val
                        else:
                            self[name]['data'] = 0.0 * psin
                            missing += [name]

                    for name in missing:
                        printw(f"pFile.from_omas: {name} could not be defined")

        for key in list(self.keys()):
            if key != 'N Z A':
                if len(self[key]['data']) == 1:
                    del self[key]
                else:
                    self[key]['derivative'] = deriv(self[key]['psinorm'], self[key]['data'])

        # Hahm-Burrell shearing rate
        #        self['omghb'] = OMFITncData()
        #        self['omghb']['data'] = 1e-3*(R*Bp)**2*self['omgeb']['derivative']
        #        self['omghb']['data'] /= abs(gEQDSK['SIBRY']-gEQDSK['SIMAG'])*np.sqrt(Bt**2 + Bp**2)
        #        self['omghb']['description'] = self.descriptions['omghb']
        #        self['omghb']['psinorm'] = psin
        #        self['omghb']['units'] = self.units['omghb']
        #        self['omghb']['derivative'] = deriv(self['omghb']['psinorm'], self['omghb']['data'])
        #        printw('Hahm-Burrell shearing rate not written to pFile')

        return self


# OMAS extra_structures
# omega0 is kept for backward compatibility
_extra_structures = {
    'core_profiles': {
        "core_profiles.profiles_1d[:].omega0": {
            "coordinates": ["core_profiles.profiles_1d[:].grid.rho_tor_norm"],
            "data_type": "FLT_1D",
            "documentation": "ExB plasma rotation",
            "full_path": "core_profiles/profiles_1d(itime)/omega0(:)",
            "data_type": "FLT_1D",
            "units": "rad/s",
            "cocos_signal": "TOR",
        },
        "core_profiles.profiles_1d[:].electrons.rotation.perpendicular": {
            "coordinates": ["core_profiles.profiles_1d[:].grid.rho_tor_norm"],
            "data_type": "FLT_1D",
            "documentation": "electron perpendicular VxB rotation",
            "full_path": "core_profiles/profiles_1d(itime)/electrons/rotation/perpendicular(:)",
            "data_type": "FLT_1D",
            "units": "rad/s",
            "cocos_signal": "TOR",
        },
        "core_profiles.profiles_1d[:].electrons.rotation.diamagnetic": {
            "coordinates": ["core_profiles.profiles_1d[:].grid.rho_tor_norm"],
            "data_type": "FLT_1D",
            "documentation": "electron diamagnetic rotation",
            "full_path": "core_profiles/profiles_1d(itime)/electrons/rotation/diamagnetic(:)",
            "data_type": "FLT_1D",
            "units": "rad/s",
            "cocos_signal": "TOR",
        },
        "core_profiles.profiles_1d[:].ion[:].rotation.perpendicular": {
            "coordinates": ["core_profiles.profiles_1d[:].grid.rho_tor_norm"],
            "data_type": "FLT_1D",
            "documentation": "ion perpendicular VxB rotation",
            "full_path": "core_profiles/profiles_1d(itime)/ion(i1)/rotation/perpendicular(:)",
            "data_type": "FLT_1D",
            "units": "rad/s",
            "cocos_signal": "TOR",
        },
        "core_profiles.profiles_1d[:].ion[:].rotation.diamagnetic": {
            "coordinates": ["core_profiles.profiles_1d[:].grid.rho_tor_norm"],
            "data_type": "FLT_1D",
            "documentation": "ion diamagnetic rotation",
            "full_path": "core_profiles/profiles_1d(itime)/ion(i1)/rotation/diamagnetic(:)",
            "data_type": "FLT_1D",
            "units": "rad/s",
            "cocos_signal": "TOR",
        },
        "core_profiles.profiles_1d[:].ion[:].rotation.parallel_stream_function": {
            "coordinates": ["core_profiles.profiles_1d[:].grid.rho_tor_norm"],
            "data_type": "FLT_1D",
            "documentation": "ion parallel stream function",
            "full_path": "core_profiles/profiles_1d(itime)/ion(i1)/rotation/parallel_stream_function(:)",
            "data_type": "FLT_1D",
            "units": "rad/s"
            # no COCOS transformation
        },
    }
}
omas.omas_utils._structures = {}
omas.omas_utils._structures_dict = {}
if not hasattr(omas.omas_utils, '_extra_structures'):
    omas.omas_utils._extra_structures = {}
for _ids in _extra_structures:
    for _item in _extra_structures[_ids]:
        _extra_structures[_ids][_item]['lifecycle_status'] = 'omfit'
    omas.omas_utils._extra_structures.setdefault(_ids, {}).update(_extra_structures[_ids])
